UNIT actor;
(* Duke of Dragonfear: Deathland.
 *	An action game for the 1st Pascal Game Development Mini Contest.
 *
 * An "actor" is a game item that has a rol in the game, i.e. the player,
 * enemies, etc.
 *
 * This unit implements classes used as base for these kind of objects.
 *)
(*
  Copyright (c) 2011 Guillermo Martínez J.

  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software
    in a product, an acknowledgment in the product documentation would be
    appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
    distribution.
*)

INTERFACE

USES
  allegro,
  contnrs;

CONST
(* Mnemonics for directions in picture list. *)
  RS_W  = 1;
  RS_NW = 2;
  RS_N  = 3;
  RS_NE = 4;
  RS_E  = 5;
  RS_SE = 6;
  RS_S  = 7;
  RS_SW = 8;
(* Direction groups. *)
  RS_G_N = [RS_NW, RS_N, RS_NE];
  RS_G_S = [RS_SW, RS_S, RS_SE];
  RS_G_E = [RS_NE, RS_E, RS_SE];
  RS_G_W = [RS_NW, RS_W, RS_SW];
(* Mnemonics for animations in picture list. *)
  RS_STAND      = 1;
  RS_WALK_BEGIN = 1;
  RS_WALK_END   = 4;
  RS_ATTACK_1   = 5;
  RS_ATTACK_2   = 6;
  RS_DEFENSE    = 6;
  RS_PAIN       = 7;
  RS_DEAD       = 8;

TYPE
(* Loads and helps managing all animations of a game character. *)
  TCharacterPictureList = CLASS (TObject)
  PRIVATE
  (* The pictures.  First index is direction, second is animation. *)
    fPictures: ARRAY [1..8, 1..8] OF AL_RLE_SPRITEptr;
  (* Returns pointer to RLE sprite. *)
    FUNCTION GetRleSprite (Dir, Act: INTEGER): AL_RLE_SPRITEptr; INLINE;
  PUBLIC
  (* Constructor.  Extracts the RLE sprites from the given bitmap.  This is an
     8x8 matrix where line is direction and column is animation.
     It copyes the data so the bitmap should be destroyed by caller. *)
    CONSTRUCTOR Create (aBitmap: AL_BITMAPptr);
  (* Releases data. *)
    DESTRUCTOR Destroy; OVERRIDE;
  (* Returns pointer to the picture. *)
    PROPERTY Picture[Dir, Act: INTEGER]: AL_RLE_SPRITEptr READ GetRleSprite; DEFAULT;
  END;



(* Mnemonics to know what the actor is doing. *)
  TActorAction = (
  (* Actor has received an attack. *)
    aaAching,
  (* Actor is attacking. *)
    aaAttacking,
  (* Actor is dead. *)
    aaDead,
  (* Actor is dodgeing an attack. *)
    aaDodge,
  (* Actor is scentry, looking for enemies. *)
    aaScentry,
  (* Actor does nothing. *)
    aaStop,
  (* Actor is waiting the best moment to attack. *)
    aaWaiting,
  (* Actor is walking. *)
    aaWalking
  );



(* Base for actors. *)
  TActor = CLASS (TObject)
  PROTECTED
  (* Wat is it.  It should be an Allegro's Id created by "AL_ID" funciton. *)
    fTypeId: INTEGER;
  (* Health points.  If it's 0 or less, it's dead. *)
    fHealthPoints: INTEGER;
  (* Index of the sprite in the sprite engine. *)
    fSprIndex: INTEGER;
  PUBLIC
  (* Creates the object. *)
    CONSTRUCTOR Create (aSprIndex: INTEGER = -1);
  (* Updates the actor. *)
    PROCEDURE Update; VIRTUAL;
  (* Method to tell to the actor that it was collided with another actor. *)
    PROCEDURE CollidedWidth (Actor: TActor); VIRTUAL;
  (* Method to tell to the actor that another one has attacked him. *)
    PROCEDURE Attacked (Strength: INTEGER); VIRTUAL;
  (* Actor type. *)
    PROPERTY TypeId: INTEGER READ fTypeId;
  (* Health points.  If it's 0 or less, it's dead. *)
    PROPERTY HealthPoints: INTEGER READ fHealthPoints;
  (* Index of the sprite. *)
    PROPERTY SpriteIndex: INTEGER READ fSprIndex;
  END;



(* Manages actors. *)
  TActorList = CLASS (TObject)
  PRIVATE
  (* The enemies. *)
    fEnemies: TFPObjectList;
  (* To get list properties. *)
    FUNCTION ActorCount: INTEGER; INLINE;
    FUNCTION GetActor (Ndx: INTEGER): TActor; INLINE;
  PUBLIC
  (* Constructor. *)
    CONSTRUCTOR Create;
  (* Destructor.  It desroys enemies too. *)
    DESTRUCTOR Destroy; OVERRIDE;
  (* Adds an actor to the list.
     @returns(actor index.) *)
    FUNCTION Add (aActor: TActor) :INTEGER; INLINE;
  (* Extracts actor from list without destroy it.
     @returns(reference to the actor.) *)
    FUNCTION Extract (Ndx: INTEGER): TActor; INLINE;
  (* Destroys actor. *)
    PROCEDURE Remove (Ndx: INTEGER); INLINE;
  (* Empties the list destroying the enemies. *)
    PROCEDURE Clear; INLINE;
  (* Updates the enemies.  That is, calls "Update" for each actor. *)
    PROCEDURE Update; INLINE;
  (* Actor number. *)
    PROPERTY Count: INTEGER READ ActorCount;
  (* Returns actor by index. *)
    PROPERTY Actor[Ndx: INTEGER]: TActor READ GetActor;
  END;



TYPE
(* Helps to store sound samples. *)
  TSoundList = RECORD
    Happy,
    Attack,
    Dodge,
    Pain: AL_SAMPLEptr;
  END;



(* Actor that represent a character. *)
  TCharacterActor = CLASS (TActor)
  PROTECTED
  (* What's doing. *)
    fStatus: TActorAction;
  (* Reference to picture list. *)
    fPictures: TCharacterPictureList;
  (* Reference to sound samples. *)
    fSamples: TSoundList;
  (* Index of the sprite used to check "weapon" colision. *)
    fWeaponSprIndex: INTEGER;
  (* Sprite direction. *)
    fDirection: INTEGER;
  (* A counter with different uses depending on the character status. *)
    fCounter: INTEGER;

  (* Inits an attack.  It puts the values of the weapon sprite, the status and
     the counter. *)
    PROCEDURE InitAttack; INLINE;
  (* Inits the "hurting" status.  It puts the sprite, status and counter
     values, and decreases health points. *)
    PROCEDURE InitHurt (Strength: INTEGER); INLINE;
  PUBLIC
  (* Creates the object, assigning the picture list.  The picture list is a
     reference, not an assignation. *)
    CONSTRUCTOR Create (aPictureList: TCharacterPictureList; aSampleList: TSoundList; aSprIndex: INTEGER); VIRTUAL;
  (* Finalizes the character. *)
    DESTRUCTOR Destroy; OVERRIDE;
  (* Index of the weapon sprite. *)
    PROPERTY WeaponSpriteIndex: INTEGER READ fWeaponSprIndex;
  (* Status. *)
    PROPERTY Status: TActorAction READ fStatus;
  END;

VAR
(* Identifer for different actor types. *)
  atPlayer, atEnemy: INTEGER;

IMPLEMENTATION

USES
  Sprites, TileMap,
  sysutils;

(*************************
 * TCharacterPictureList *
 *************************)

(* Returns pointer to RLE sprite. *)
  FUNCTION TCharacterPictureList.GetRleSprite (Dir, Act: INTEGER): AL_RLE_SPRITEptr;
  BEGIN
    GetRleSprite := fPictures[Dir, Act];
  END;



(* Constructor.  Extracts the RLE sprites from the given bitmap.  This is an
   8x8 matrix where line is direction and column is animation.
   It copyes the data so the bitmap should be destroyed by caller. *)
  CONSTRUCTOR TCharacterPictureList.Create (aBitmap: AL_BITMAPptr);
  VAR
    Action, Direction, SizeW, SizeH: INTEGER;
    SubBitmap: AL_BITMAPptr;
  BEGIN
    INHERITED Create;
    SizeW := aBitmap^.w DIV 8;
    SizeH := aBitmap^.h DIV 8;
    FOR Direction := RS_W TO RS_SW DO
      FOR Action := RS_STAND TO RS_DEAD DO
      BEGIN
	TRY
	  SubBitmap := al_create_sub_bitmap (aBitmap,
	    (Action - 1) * SizeW, (Direction - 1) * SizeH,
	    SizeW, SizeH);
	  fPictures[Direction, Action] := al_get_rle_sprite (SubBitmap);
	FINALLY
	  al_destroy_bitmap (SubBitmap);
	END;
      END;
  END;



(* Releases data. *)
  DESTRUCTOR TCharacterPictureList.Destroy;
  VAR
    Action, Direction: INTEGER;
  BEGIN
    FOR Direction := RS_W TO RS_SW DO
      FOR Action := RS_STAND TO RS_DEAD DO
	IF fPictures[Direction, Action] <> NIL THEN
	  al_destroy_rle_sprite (fPictures[Direction, Action]);
    INHERITED Destroy;
  END;



(**********
 * TActor *
 **********)

(* Creates the object. *)
  CONSTRUCTOR TActor.Create (aSprIndex: INTEGER);
  BEGIN
    INHERITED Create;
    IF aSprIndex < 0 THEN
      RAISE Exception.Create ('Can''t add actor to the sprite-list.  More sprites needed!');
    fSprIndex := aSprIndex
  END;



(* Updates the actor. *)
  PROCEDURE TActor.Update;
  BEGIN
    { By default it doesn't do nothing. }
  END;



(* Method to tell to the actor that it was collided with another actor. *)
  PROCEDURE TActor.CollidedWidth (Actor: TActor);
  BEGIN
    { By default it doesn't do nothing. }
  END;



(* Method to tell to the actor that another one has attack him. *)
  PROCEDURE TActor.Attacked (Strength: INTEGER);
  BEGIN
    { By default it doesn't do nothing. }
  END;



(**************
 * TActorList *
 **************)

(* Returns actor counter. *)
  FUNCTION TActorList.ActorCount: INTEGER;
  BEGIN
    ActorCount := fEnemies.Count;
  END;



(* Returns the actor. *)
  FUNCTION TActorList.GetActor (Ndx: INTEGER): TActor;
  BEGIN
    GetActor := fEnemies.Items[Ndx] AS TActor;
  END;



(* Constructor. *)
  CONSTRUCTOR TActorList.Create;
  BEGIN
    INHERITED Create;
    fEnemies := TFPObjectList.Create;
  END;



(* Destructor. *)
  DESTRUCTOR TActorList.Destroy;
  BEGIN
    fEnemies.Free;
    INHERITED Destroy;
  END;



(* Adds an actor to the list.
   @returns(actor index.) *)
  FUNCTION TActorList.Add (aActor: TActor) :INTEGER;
  BEGIN
    Add := fEnemies.Add (aActor);
  END;



(* Extracts actor from list without destroy it.
   @returns(reference to the actor.) *)
  FUNCTION TActorList.Extract (Ndx: INTEGER): TActor;
  BEGIN
    Extract := fEnemies.Extract (fEnemies.Items [Ndx]) AS TActor;
  END;



(* Destroys actor. *)
  PROCEDURE TActorList.Remove (Ndx: INTEGER);
  BEGIN
    fEnemies.Delete (Ndx);
  END;



(* Empties the list destroying the enemies. *)
  PROCEDURE TActorList.Clear;
  BEGIN
    fEnemies.Clear;
  END;


(* Updates the enemies.  That is, calls "Update" for each actor. *)
  PROCEDURE TActorList.Update;
  VAR
    Ndx: INTEGER;
    NeedsPack: BOOLEAN;
  BEGIN
    NeedsPack := FALSE;
    FOR Ndx := 0 TO fEnemies.Count - 1 DO
      IF fEnemies.Items[Ndx] <> NIL THEN
	(fEnemies.Items[Ndx] AS TActor).Update
      ELSE
	NeedsPack := TRUE;
    IF NeedsPack THEN
      fEnemies.Pack;
  END;



(*******************
 * TCharacterActor *
 *******************)

(* Inits an attack.  It puts the values for character sprite, weapon sprite,
   status and counter. *)
  PROCEDURE TCharacterActor.InitAttack;
  BEGIN
    Sprites.PlaySoundSample (fSprIndex, fSamples.Attack);
    fStatus := aaAttacking;
    fCounter := 0;
    SpriteList[fSprIndex].Picture := fPictures.Picture[fDirection, RS_ATTACK_1];
    IF fDirection IN RS_G_N THEN
      SpriteList[fWeaponSprIndex].Y := SpriteList[fSprIndex].Y - (TSIZE DIV 2)
    ELSE
      SpriteList[fWeaponSprIndex].Y := SpriteList[fSprIndex].Y;
    IF fDirection = RS_W THEN
    BEGIN
      SpriteList[fWeaponSprIndex].X := SpriteList[fSprIndex].X - TSIZE;
      SpriteList[fWeaponSprIndex].Y := SpriteList[fSprIndex].Y - (TSIZE DIV 4);
    END
    ELSE IF fDirection IN [RS_NW, RS_SW] THEN
      SpriteList[fWeaponSprIndex].X := SpriteList[fSprIndex].X - (TSIZE DIV 2)
    ELSE IF fDirection = RS_E THEN
    BEGIN
      SpriteList[fWeaponSprIndex].X := SpriteList[fSprIndex].X + TSIZE;
      SpriteList[fWeaponSprIndex].Y := SpriteList[fSprIndex].Y - (TSIZE DIV 4);
    END
    ELSE IF fDirection IN [RS_NE, RS_SE] THEN
      SpriteList[fWeaponSprIndex].X := SpriteList[fSprIndex].X + (TSIZE DIV 2)
    ELSE
      SpriteList[fWeaponSprIndex].X := SpriteList[fSprIndex].X;
  END;



(* Inits the "hurting" status.  It puts the sprite, status and counter
   values, and decreases health points. *)
  PROCEDURE TCharacterActor.InitHurt (Strength: INTEGER);
  BEGIN
    IF NOT (fStatus IN [aaDead, aaAching]) THEN
    BEGIN
      Sprites.PlaySoundSample (fSprIndex, fSamples.Pain);
      fStatus := aaAching;
      fCounter := 0;
      DEC (fHealthPoints, Strength);
      SpriteList[fSprIndex].Picture := fPictures.Picture[fDirection, RS_PAIN];
    END;
  END;



(* Creates the object, assigning the RLE  objects.  The RLE list is a
   reference, not an assignation. *)
  CONSTRUCTOR TCharacterActor.Create (aPictureList: TCharacterPictureList; aSampleList: TSoundList; aSprIndex: INTEGER);
  BEGIN
    INHERITED Create (aSprIndex);
    fStatus := aaStop;
    fPictures := aPictureList;
    fSamples := aSampleList;
    fDirection := RS_S;
    fCounter := 0;
    fWeaponSprIndex := -1;
  { All actors has the same sprite size. }
    SpriteList[fSprIndex].OffX :=  -95;
    SpriteList[fSprIndex].OffY := -127;
  END;



(* Finalizes the character. *)
  DESTRUCTOR TCharacterActor.Destroy;
  BEGIN
  { Deactivates the weapon sprite. }
    IF fWeaponSprIndex > -1 THEN
      SpriteList[fWeaponSprIndex].Y := SPR_Y_OUT;
    INHERITED Destroy;
  END;



INITIALIZATION
  atPlayer := AL_ID ('PLYR');
  atEnemy  := AL_ID ('ENMY');
END.
