UNIT Enemies;
(* Duke of Dragonfear: Deathland.
 *	An action game for the 1st Pascal Game Development Mini Contest.
 *
 * Implements the enemies.
 *)
(*
  Copyright (c) 2011 Guillermo Martínez J.

  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software
    in a product, an acknowledgment in the product documentation would be
    appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
    distribution.
*)

INTERFACE

USES
  actor;

TYPE
(* Generic enemy. *)
  TEnemy = CLASS (TCharacterActor)
  PROTECTED
  (* Chances of dodge a player attack are 1/fDodge. *)
    fDodge: INTEGER;
  (* Check if player is attacking. *)
    FUNCTION CheckPlayerAttack: BOOLEAN; INLINE;
  (* Looks for an objective to attack. *)
    FUNCTION LocateObjective (MaxDistance: LONGINT): TCharacterActor;
  PUBLIC
  (* Puts the enemy in place. *)
    CONSTRUCTOR Create (aPictureList: TCharacterPictureList; aSampleList: TSoundList; aSprIndex, Px, Py: INTEGER); OVERLOAD;
  END;



(* The Zombie actor. *)
  TZombie = CLASS (TEnemy)
  PUBLIC
  (* Creates the zombie. *)
    CONSTRUCTOR Create (aSprIndex, Px, Py: INTEGER); OVERLOAD;
  (* The zombie inteligence... or someting. *)
    PROCEDURE Update; OVERRIDE;
  END;



(* The Skeleton actor. *)
  TSkeleton = CLASS (TEnemy)
  PUBLIC
  (* Creates the zombie. *)
    CONSTRUCTOR Create (aSprIndex, Px, Py: INTEGER); OVERLOAD;
  (* The zombie inteligence... or someting. *)
    PROCEDURE Update; OVERRIDE;
  END;

CONST
  ZOMBIE_STRENGTH = 1;
  SKELETON_STRENGTH = 1;

VAR
{ List of enemies. }
  EnemyList: TActorList;

IMPLEMENTATION

USES
  gamedata, Geometry, graphsys, Player, Sprites, TileMap;

CONST
  ZOMBIE_MAX_CAN_SEE_DISTANCE = 250;
  ZOMBIE_MAX_CAN_SEE_DISTANCE2 = ZOMBIE_MAX_CAN_SEE_DISTANCE * ZOMBIE_MAX_CAN_SEE_DISTANCE;

  ZOMBIE_MAX_CAN_ATTACK_DISTANCE = TSIZE;
  ZOMBIE_MAX_CAN_ATTACK_DISTANCE2 = ZOMBIE_MAX_CAN_ATTACK_DISTANCE * ZOMBIE_MAX_CAN_ATTACK_DISTANCE;

  SKELETON_MAX_CAN_SEE_DISTANCE = 300;
  SKELETON_MAX_CAN_SEE_DISTANCE2 = SKELETON_MAX_CAN_SEE_DISTANCE * SKELETON_MAX_CAN_SEE_DISTANCE;

  SKELETON_MAX_CAN_ATTACK_DISTANCE = TSIZE;
  SKELETON_MAX_CAN_ATTACK_DISTANCE2 = SKELETON_MAX_CAN_ATTACK_DISTANCE * SKELETON_MAX_CAN_ATTACK_DISTANCE;



(**********
 * TEnemy *
 **********)

(* Check if player is attacking. *)
  FUNCTION TEnemy.CheckPlayerAttack: BOOLEAN;
  VAR
    Ndx, NumAttacks: INTEGER;
  BEGIN
    CheckPlayerAttack := FALSE;
    NumAttacks := 0;
    FOR Ndx := 0 TO PlayerList.Count -1 DO
      IF ((PlayerList.Actor[Ndx] AS TCharacterActor).Status = aaAttacking)
      AND Sprites.Collision ((PlayerList.Actor[Ndx] AS TCharacterActor).WeaponSpriteIndex, fSprIndex)
      THEN
	INC (NumAttacks);
    IF NumAttacks > 0 THEN
    BEGIN
      IF Random (fDodge) = 1 THEN
      BEGIN
	Sprites.PlaySoundSample (fSprIndex, fSamples.Dodge);
	fStatus := aaDodge;
	fCounter := 0;
	SpriteList[fSprIndex].Picture := fPictures.Picture[fDirection, RS_DEFENSE];
      END
      ELSE
	InitHurt (PLAYER_STRENGTH * NumAttacks);
      CheckPlayerAttack := TRUE;
    END;
  END;



(* Looks for an objective to attack. *)
  FUNCTION TEnemy.LocateObjective (MaxDistance: LONGINT): TCharacterActor;
  VAR
    Distance, DistanceObjective: LONGINT;
    Ndx: INTEGER;
    TmpObjective: TCharacterActor;
  BEGIN
  { Initialices values to "didn't find yet". }
    DistanceObjective := $7FFFFFFF;
    LocateObjective := NIL;
  { Look for the objective. }
    FOR Ndx := 0 TO PlayerList.Count -1 DO
    BEGIN
      TmpObjective := PlayerList.Actor[Ndx] AS TCharacterActor;
      IF TmpObjective.Status <> aaDead THEN
      BEGIN
	Distance := Distance2 (
	  SpriteList[TmpObjective.SpriteIndex].X, SpriteList[TmpObjective.SpriteIndex].Y,
	  SpriteList[fSprIndex].X, SpriteList[fSprIndex].Y);
	IF (Distance <= MaxDistance) AND (Distance < DistanceObjective) THEN
	  IF (TmpObjective = NIL) OR  (Random (100) > 1) THEN
	  BEGIN
	    DistanceObjective := Distance;
	    LocateObjective := TmpObjective;
	  END;
      END;
    END;
  END;



(* Puts the enemy in place. *)
  CONSTRUCTOR TEnemy.Create (aPictureList: TCharacterPictureList; aSampleList: TSoundList; aSprIndex, Px, Py: INTEGER);
  BEGIN
    INHERITED Create (aPictureList, aSampleList, aSprIndex);
    fTypeId := atEnemy;
    fWeaponSprIndex := Sprites.NextFreeSprite;
    fStatus := aaScentry;
    fDirection := Random (RS_SW) + 1;
    WITH SpriteList[fSprIndex] DO
    BEGIN
      X := Px * TSIZE;
      Y := Py * TSIZE;
      Picture := fPictures.Picture[fDirection, RS_STAND];
    END;
  END;



(***********
 * TZombie *
 ***********)

(* Creates the zombie. *)
  CONSTRUCTOR TZombie.Create (aSprIndex, Px, Py: INTEGER);
  BEGIN
    INHERITED Create (gamedata.ZombiePictures, gamedata.ZombieSamples, aSprIndex, Px, Py);
    fHealthPoints := 4;
    fDodge := 5;
  END;



(* The zombie inteligence... or someting. *)
  PROCEDURE TZombie.Update;
  VAR
    Dir, Ndx: INTEGER;
    Objective: TCharacterActor;

  (* Helper function to know if it is near the player. *)
    FUNCTION IsNearPlayer: BOOLEAN; INLINE;
    BEGIN
      Objective := LocateObjective (ZOMBIE_MAX_CAN_SEE_DISTANCE2);
      IsNearPlayer := Objective <> NIL;
    END;

  BEGIN
  { Each status has different process. }
    CASE fStatus OF
      aaAching:
      BEGIN INC (fCounter);
	IF fCounter >= 40 THEN
	BEGIN
	  IF fHealthPoints > 0 THEN
	    fStatus := aaScentry
	  ELSE BEGIN
	    SpriteList[fSprIndex].Picture := fPictures.Picture[fDirection, RS_DEAD];
	    fStatus := aaDead;
	  END;
	END;
      END;
      aaAttacking: BEGIN
	INC (fCounter);
	IF fCounter = 20 THEN
	  SpriteList[fWeaponSprIndex].Y := SPR_Y_OUT
	ELSE IF fCounter = 30 THEN
	  SpriteList[fSprIndex].Picture := fPictures.Picture[fDirection, RS_STAND]
	ELSE IF fCounter >= 100 THEN
	BEGIN
	  fCounter := 0;
	  fStatus := aaWaiting;
	END
	ELSE IF (fCounter > 10) THEN
	  FOR Ndx := 0 TO PlayerList.Count - 1 DO
	    IF Sprites.Collision (fWeaponSprIndex, PlayerList.Actor[Ndx].SpriteIndex) THEN
	      PlayerList.Actor[Ndx].Attacked (ZOMBIE_STRENGTH);
      END;
      aaDodge: BEGIN
	INC (fCounter);
	IF fCounter = 30 THEN
	BEGIN
	  SpriteList[fSprIndex].Picture := fPictures.Picture[fDirection, RS_STAND];
	  fStatus := aaWaiting;
	END;
      END;
      aaScentry:
      IF NOT CheckPlayerAttack THEN
      BEGIN
	Dir := 0;
	IF IsNearPlayer THEN
	BEGIN
	  Sprites.PlaySoundSample (fSprIndex, fSamples.Happy);
	  fStatus := aaWaiting;
	  Dir := LookAt (
	    SpriteList[fSprIndex].X, SpriteList[fSprIndex].Y,
	    SpriteList[Objective.SpriteIndex].X, SpriteList[Objective.SpriteIndex].Y
	    );
	END
	ELSE BEGIN
	{ Random walking. }
	  IF Random (20000) < 1 THEN
	    Dir := Random (8) + 1;
	  IF Random (2000) < 1 THEN
	    fStatus := aaWalking;
	END;
	IF (Dir <> 0) AND (fDirection <> Dir) THEN
	BEGIN
	  fDirection := Dir;
	  fCounter := 0;
	  SpriteList[fSprIndex].Picture := fPictures.Picture[Dir, RS_STAND];
	END;
      END;
      aaWaiting:
      IF NOT CheckPlayerAttack THEN
      BEGIN
	INC (fCounter);
	IF fCounter >= 40 THEN
	BEGIN
	  fCounter := 0;
	{ Check if it's near the player. }
	  IF IsNearPlayer AND (Objective.Status <> aaDead) THEN
	  BEGIN
	    fDirection := LookAt (
	      SpriteList[fSprIndex].X, SpriteList[fSprIndex].Y,
	      SpriteList[Objective.SpriteIndex].X, SpriteList[Objective.SpriteIndex].Y
	    );
	    IF Distance2 (
	      SpriteList[Objective.SpriteIndex].X, SpriteList[Objective.SpriteIndex].Y,
	      SpriteList[fSprIndex].X, SpriteList[fSprIndex].Y) <= ZOMBIE_MAX_CAN_ATTACK_DISTANCE2
	    THEN
	    BEGIN
	      IF Random (10) > 3 THEN
	      BEGIN
		InitAttack;
	      END;
	    END
	    ELSE IF Random (10) > 0 THEN
	      fStatus := aaWalking
	    ELSE
	      fCounter := 0;
	  END
	  ELSE IF Random (3) = 1 THEN
	    fStatus := aaWalking
	  ELSE
	    fStatus := aaScentry;
	END;
      END;
      aaWalking:
      IF NOT CheckPlayerAttack THEN
      BEGIN
	INC (fCounter);
	IF fCounter MOD 2 = 1 THEN
	BEGIN
	{ Checks walking direction. }
	  IF (fDirection IN RS_G_W)
	  AND NOT CheckLeftCollision (fSprIndex) THEN
	    DEC (SpriteList[fSprIndex].x);
	  IF (fDirection IN RS_G_E)
	  AND NOT CheckRightCollision (fSprIndex) THEN
	    INC (SpriteList[fSprIndex].x);
	  IF (fDirection IN RS_G_N)
	  AND NOT CheckCeilCollision (fSprIndex) THEN
	    DEC (SpriteList[fSprIndex].y);
	  IF (fDirection IN RS_G_S)
	  AND NOT CheckGroundCollision (fSprIndex) THEN
	    INC (SpriteList[fSprIndex].y);
	END;
	IF fCounter < 40 THEN
	  SpriteList[fSprIndex].Picture := fPictures.Picture[fDirection, (fCounter DIV 10) + 1]
	ELSE BEGIN
	  fStatus := aaWaiting;
	  fCounter := 0;
	END;
      END;
    END;
  END;



(*************
 * TSkeleton *
 *************)

(* Creates the zombie. *)
  CONSTRUCTOR TSkeleton.Create (aSprIndex, Px, Py: INTEGER);
  BEGIN
    INHERITED Create (gamedata.SkeletonPictures, gamedata.SkeletonSamples, aSprIndex, Px, Py);
    fHealthPoints := 4;
    fDodge := 5;
  END;



(* The zombie inteligence... or someting. *)
  PROCEDURE TSkeleton.Update;
  VAR
    Dir, Ndx: INTEGER;
    Objective: TCharacterActor;

  (* Helper function to know if it is near the player. *)
    FUNCTION IsNearPlayer: BOOLEAN; INLINE;
    BEGIN
      Objective := LocateObjective (SKELETON_MAX_CAN_SEE_DISTANCE2);
      IsNearPlayer := Objective <> NIL;
    END;

  BEGIN
  { Each status has different process. }
    CASE fStatus OF
      aaAching:
      BEGIN INC (fCounter);
	IF fCounter >= 25 THEN
	BEGIN
	  IF fHealthPoints > 0 THEN
	    fStatus := aaScentry
	  ELSE BEGIN
	    SpriteList[fSprIndex].Picture := fPictures.Picture[fDirection, RS_DEAD];
	    fStatus := aaDead;
	  END;
	END;
      END;
      aaAttacking: BEGIN
	INC (fCounter);
	IF fCounter = 20 THEN
	  SpriteList[fWeaponSprIndex].Y := SPR_Y_OUT
	ELSE IF fCounter = 25 THEN
	  SpriteList[fSprIndex].Picture := fPictures.Picture[fDirection, RS_STAND]
	ELSE IF fCounter >= 40 THEN
	BEGIN
	  fCounter := 0;
	  fStatus := aaWaiting;
	END
	ELSE IF (fCounter > 5) THEN
	  FOR Ndx := 0 TO PlayerList.Count - 1 DO
	    IF Sprites.Collision (fWeaponSprIndex, PlayerList.Actor[Ndx].SpriteIndex) THEN
	      PlayerList.Actor[Ndx].Attacked (SKELETON_STRENGTH);
      END;
      aaDodge: BEGIN
	INC (fCounter);
	IF fCounter >= 15 THEN
	BEGIN
	  SpriteList[fSprIndex].Picture := fPictures.Picture[fDirection, RS_STAND];
	  fStatus := aaWaiting;
	END;
      END;
      aaScentry:
      IF NOT CheckPlayerAttack THEN
      BEGIN
	Dir := 0;
	IF IsNearPlayer THEN
	BEGIN
	  Sprites.PlaySoundSample (fSprIndex, fSamples.Happy);
	  fStatus := aaWaiting;
	  Dir := LookAt (
	    SpriteList[fSprIndex].X, SpriteList[fSprIndex].Y,
	    SpriteList[Objective.SpriteIndex].X, SpriteList[Objective.SpriteIndex].Y
	  );
	END
	ELSE BEGIN
	{ Random walking. }
	  IF Random (20000) < 1 THEN
	    Dir := Random (8) + 1;
	  IF Random (2000) < 1 THEN
	    fStatus := aaWalking;
	END;
	IF (Dir <> 0) AND (fDirection <> Dir) THEN
	BEGIN
	  fDirection := Dir;
	  fCounter := 0;
	  SpriteList[fSprIndex].Picture := fPictures.Picture[Dir, RS_STAND];
	END;
      END;
      aaWaiting:
      IF NOT CheckPlayerAttack THEN
      BEGIN
	INC (fCounter);
	fCounter := 0;
      { Check if it's near the player. }
	IF IsNearPlayer THEN
	BEGIN
	  fDirection := LookAt (
	    SpriteList[fSprIndex].X, SpriteList[fSprIndex].Y,
	    SpriteList[Objective.SpriteIndex].X, SpriteList[Objective.SpriteIndex].Y
	  );
	  IF Distance2 (
	    SpriteList[Objective.SpriteIndex].X, SpriteList[Objective.SpriteIndex].Y,
	    SpriteList[fSprIndex].X, SpriteList[fSprIndex].Y) <= SKELETON_MAX_CAN_ATTACK_DISTANCE2
	  THEN
	  BEGIN
	    IF Random (10) > 3 THEN
	    BEGIN
	      InitAttack;
	    END;
	  END
	  ELSE IF Random (10) > 0 THEN
	    fStatus := aaWalking
	  ELSE
	    fCounter := 0;
	END
	ELSE IF Random (3) = 1 THEN
	  fStatus := aaWalking
	ELSE
	  fStatus := aaScentry;
      END;
      aaWalking:
      IF NOT CheckPlayerAttack THEN
      BEGIN
	INC (fCounter);
      { Checks walking direction. }
	IF (fDirection IN RS_G_W)
	AND NOT CheckLeftCollision (fSprIndex) THEN
	  DEC (SpriteList[fSprIndex].x);
	IF (fDirection IN RS_G_E)
	AND NOT CheckRightCollision (fSprIndex) THEN
	  INC (SpriteList[fSprIndex].x);
	IF (fDirection IN RS_G_N)
	AND NOT CheckCeilCollision (fSprIndex) THEN
	  DEC (SpriteList[fSprIndex].y);
	IF (fDirection IN RS_G_S)
	AND NOT CheckGroundCollision (fSprIndex) THEN
	  INC (SpriteList[fSprIndex].y);
	IF fCounter < 40 THEN
	  SpriteList[fSprIndex].Picture := fPictures.Picture[fDirection, (fCounter DIV 10) + 1]
	ELSE BEGIN
	  fStatus := aaWaiting;
	  fCounter := 0;
	END;
      END;
    END;
  END;

END.
