UNIT GameData;
(* Duke of Dragonfear: Deathland.
 *	An action game for the 1st Pascal Game Development Mini Contest.
 *
 * Unit to store and manage the game data.
 *)
(*
  Copyright (c) 2011 Guillermo Martínez J.

  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software
    in a product, an acknowledgment in the product documentation would be
    appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
    distribution.
*)

INTERFACE

USES
  actor,
  allegro;

VAR
(* Tileset. *)
  DungeonTileset: AL_BITMAPptr;
(* Actors RLE sprites. *)
  PlayerPictures, TorchManPictures,
  SkeletonPictures, ZombiePictures: TCharacterPictureList;
(* GUI stuff. *)
  HP_FullBitmap, HP_EmptyBitmap: AL_BITMAPptr;
(* Sound samples. *)
  PlayerSamples, SkeletonSamples, ZombieSamples: TSoundList;
(* Transparent color used by sprites. *)
  MaskColor: LONGINT;



(* Loads game data. *)
  PROCEDURE Load;

(* Releases game data. *)
  PROCEDURE Destroy;

IMPLEMENTATION

USES
  Config,
  alfile, sysutils;

{$INCLUDE sound.inc}
{$INCLUDE gui.inc}

VAR
  TilesetData: AL_DATAFILEptr;
  GUIData: AL_DATAFILEptr;
  SoundData: AL_DATAFILEptr;

(* Loads game data. *)
  PROCEDURE Load;

  (* Helper to load datafile. *)
    FUNCTION LoadDatafile (FileName: STRING): AL_DATAFILEptr;
    BEGIN
      IF NOT FileExists (DataDirectory + FileName) THEN
	RAISE Exception.Create ('Can''t find file ' + FileName);
      LoadDatafile := al_load_datafile (DataDirectory + FileName);
    END;

  (* Since we will cut the sprite-bitmaps into smaller pieces, and the bitmaps
     are pretty big, we load each bitmap, cut it in the picture list and
     release de bitmap.  This way we use less memory. *)
    FUNCTION LoadPictures (FileName, BmpName: STRING): TCharacterPictureList;
    VAR
      DataObject: AL_DATAFILE_OBJECTptr;
    BEGIN
      IF NOT FileExists (DataDirectory + FileName) THEN
	RAISE Exception.Create ('Can''t find file ' + FileName);
      DataObject := al_load_datafile_object (DataDirectory + FileName, BmpName);
      IF Dataobject = NIL THEN
	RAISE Exception.Create ('Can''t find "'+BmpName+'" in file "'+FileName+'".');
      TRY
	LoadPictures := TCharacterPictureList.Create (DataObject^.dat);
      FINALLY
	al_unload_datafile_object (DataObject);
      END;
    END;

  (* Helper to assign the sound samples to the lists. *)
    PROCEDURE AssignSounds (FirstSound: INTEGER; OUT SoundList: TSoundList);
    BEGIN
      WITH SoundList DO
      BEGIN
	Attack := SoundData^[FirstSound].dat;
	Dodge  := SoundData^[FirstSound + 1].dat;
	Happy  := SoundData^[FirstSound + 2].dat;
	Pain   := SoundData^[FirstSound + 3].dat;
      END;
    END;

  BEGIN
    PlayerPictures := LoadPictures ('sprites.dat', 'PLAYER');
    TorchManPictures := LoadPictures ('sprites.dat', 'TORCHMAN');
    SkeletonPictures := LoadPictures ('sprites.dat', 'SKELETON');
    ZombiePictures := LoadPictures ('sprites.dat', 'ZOMBIE');

    TilesetData := LoadDatafile ('tileset.dat');
    DungeonTileset := TilesetData^[0].dat;

    GUIData := LoadDatafile ('gui.dat');
    HP_FullBitmap := GUIData^[HP_FULL].dat;
    HP_EmptyBitmap := GUIData^[HP_EMPTY].dat;

    SoundData := LoadDatafile ('sound.dat');
    AssignSounds (PLAYER_ATTACK, PlayerSamples);
    AssignSounds (SKELETON_ATTACK, SkeletonSamples);
    AssignSounds (ZOMBIE_ATTACK, ZombieSamples);

    MaskColor := al_bitmap_mask_color (HP_FullBitmap);
  END;



(* Releases game data. *)
  PROCEDURE Destroy;
  BEGIN
    al_unload_datafile (SoundData);
    al_unload_datafile (GUIData);
    al_unload_datafile (TilesetData);
    FreeAndNil (PlayerPictures);
    FreeAndNil (SkeletonPictures);
    FreeAndNil (ZombiePictures);
  END;

END.
