UNIT graphsys;
(* Duke of Dragonfear: Deathland.
 *	An action game for the 1st Pascal Game Development Mini Contest.
 *
 * Implements the graphic system.  At the moment it's a simple double buffer.
 *)
(*
  Copyright (c) 2011 Guillermo Martínez J.

  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software
    in a product, an acknowledgment in the product documentation would be
    appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
    distribution.
*)

INTERFACE

USES
  allegro;



(* Sets the graphic mode.  Should be called after init Allegro. *)
  PROCEDURE InitGraphicMode;



(* Returns the active page where it will draw the output. *)
  FUNCTION ActivePage: AL_BITMAPptr; INLINE;



(* Returns the visual page. *)
  FUNCTION VisualPage: AL_BITMAPptr; INLINE;



(* Swap, blit or flip pages so the active becames visible. *)
  PROCEDURE SwapVideoPages; INLINE;



(* Fade out in true-color modes isn't so easy as in VGA. 8^(
   @param(FadeSpeed How fast it does the fade.  1 needs some seconds and 256 is inmediate) *)
  PROCEDURE FadeOut (FadeSpeed: INTEGER);



(* Fades from the ActivePage to the given image.
   @param(FadeSpeed How fast it does the fade.  1 needs some seconds and 256 is inmediate) *)
  PROCEDURE FadeTo (Image: AL_BITMAPptr; FadeSpeed: INTEGER);



(* Closes the graphic mode.  MUST be called before close Allegro. *)
  PROCEDURE CloseGraphicMode;



IMPLEMENTATION

USES
  config, framectr,
  alblend, sysutils;



VAR
(* The backbuffer. *)
  fActivePage: AL_BITMAPptr;



(* Sets the grpahic mode.
 * At the moment it's a simple double-buffer system with backpage in RAM.
 * Future versions should be smart enough to enable harware page flipping,
 * triple buffer or whatever. *)
  PROCEDURE InitGraphicMode;
  VAR
    Controller: INTEGER;
  BEGIN
  { Set the driver to use. }
    IF config.FullScreen THEN
      Controller := AL_GFX_AUTODETECT_FULLSCREEN
    ELSE
      Controller := AL_GFX_AUTODETECT_WINDOWED;
  { Try to set graphic mode. }
    al_set_color_depth (al_desktop_color_depth);
    IF NOT al_set_gfx_mode (Controller, 640, 480, 0, 0) THEN
      RAISE Exception.Create ('Can''t set graphic mode!');
  { Init the back buffer. }
    IF fActivePage <> NIL THEN
      al_destroy_bitmap (fActivePage);
    fActivePage := al_create_bitmap (640, 480);
    IF fActivePage = NIL THEN
      RAISE Exception.Create ('Can''t create back buffer');
    al_set_color_conversion (AL_COLORCONV_TOTAL OR AL_COLORCONV_KEEP_TRANS);
  END;



(* Returns the active page where it will draw the output. *)
  FUNCTION ActivePage: AL_BITMAPptr;
  BEGIN
    ActivePage := fActivePage;
  END;



(* Returns the visual page. *)
  FUNCTION VisualPage: AL_BITMAPptr;
  BEGIN
    VisualPage := al_screen;
  END;



(* Swap, blit or flip pages so the active becames visible. *)
  PROCEDURE SwapVideoPages;
  BEGIN
    al_vsync;
    al_blit (fActivePage, al_screen, 0, 0, 0, 0, AL_SCREEN_W, AL_SCREEN_H);
  END;



(* Fade out in true-color modes isn't so easy as in VGA. 8^(
   @param(FadeSpeed How fast it does the fade.  1 is the slowest and 256 is the fastest) *)
  PROCEDURE FadeOut (FadeSpeed: INTEGER);
  VAR
    BlackBuffer: AL_BITMAPptr;
  BEGIN
    BlackBuffer := al_create_bitmap (ActivePage^.w, ActivePage^.h);
    IF BlackBuffer = NIL THEN
      EXIT;
    TRY
      al_clear_bitmap (BlackBuffer);
      FadeTo (BlackBuffer, FadeSpeed);
    FINALLY
      al_destroy_bitmap (BlackBuffer);
    END;
  END;



(* Fades from the ActivePage to the given image.
   @param(FadeSpeed How fast it does the fade.  1 needs some seconds and 256 is inmediate) *)
  PROCEDURE FadeTo (Image: AL_BITMAPptr; FadeSpeed: INTEGER);
  VAR
    Backup: AL_BITMAPptr;
    Level, X, Y: INTEGER;
  BEGIN
    Backup := al_create_bitmap (Image^.w, Image^.h);
    IF Backup = NIL THEN
      EXIT;
    TRY
      IF FadeSpeed < 1 THEN
	FadeSpeed := 1;
      X := (ActivePage^.w - Image^.w) DIV 2;
      Y := (ActivePage^.h - Image^.h) DIV 2;
      al_blit (ActivePage, Backup, X, Y, 0, 0, Image^.w, Image^.h);
      Level := 0;
      REPEAT
	al_set_trans_blender (0, 0, 0, 0);
	al_blit (Backup, ActivePage, 0, 0, X, Y, Image^.w, Image^.h);
	al_set_trans_blender (0, 0, 0, Level);
	al_draw_trans_sprite (ActivePage, Image, X, Y);
	SwapVideoPages;
	INC (Level, FadeSpeed);
      UNTIL Level >= 255;
      al_set_trans_blender (0, 0, 0, 0);
      al_blit (Image, ActivePage, 0, 0, X, Y, Image^.w, Image^.h);
      SwapVideoPages;
    FINALLY
      al_destroy_bitmap (Backup);
    END;
  END;



(* Closes the graphic mode.  MUST be called before close Allegro. *)
  PROCEDURE CloseGraphicMode;
  BEGIN
    al_set_gfx_mode (AL_GFX_TEXT, 0, 0, 0, 0);
    IF fActivePage <> NIL THEN
      al_destroy_bitmap (fActivePage);
    fActivePage := NIL;
  END;

END.
