UNIT Lighting;
(* Program: Duke of Dragonfear: Deathland.
 * Description: Impements the lighting system.
 *
 *		It builds a lightmap wich actually is an alpha-channel black
 *		bitmap.  When applying, the "solid" pixels are black while
 *		"transparent" ones are light.
 * Author: Ñuño Martínez <niunio@users.sourceforge.net>
 *)
(*
  Copyright (c) 2011 Guillermo Martínez J.

  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software
    in a product, an acknowledgment in the product documentation would be
    appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
    distribution.
 *)

INTERFACE

USES
  allegro;

VAR
(* The lightmap to apply to the image. *)
  LightMap: AL_BITMAPptr;



(* Initializes the lighting system. *)
  PROCEDURE Init;

(* Sets the sprite index to track by the big light. *)
  PROCEDURE SetBigLightSpriteIndex (SprIndex: INTEGER);

(* Apply the lightmap. *)
  PROCEDURE Apply (Bmp: AL_BITMAPptr; ScrollX, ScrollY: INTEGER); INLINE;

(* Releases resources. *)
  PROCEDURE ShutDown;

IMPLEMENTATION

USES
  Sprites,
  alblend;

CONST
  DARK_COMPONENT = 240;

VAR
(* The unlighted parts arent full black (0), this is the pre-calculated Color
   to use. *)
  DarkColour: LONGINT;
(* Big light lightmap. *)
  BigLight: AL_BITMAPptr;
(* The sprite index to track by the big light. *)
  BigLightSpr: INTEGER;



(* Helper to build a "spotlight". *)
  FUNCTION CreateSpotlight (Side: INTEGER): AL_BITMAPptr;
  VAR
    MaxR, r, C: INTEGER;
  BEGIN
    CreateSpotlight := al_create_bitmap (Side, Side);
    al_clear_to_color (CreateSpotlight, DarkColour);
    MaxR := (Side DIV 2);
    FOR r := 1 TO MaxR DO
    BEGIN
      C := DARK_COMPONENT - ((r * DARK_COMPONENT) DIV MaxR);
      al_circlefill (CreateSpotlight,
	Side DIV 2, Side DIV 2,
	MaxR - r,
	al_makecol (C, C, C));
    END;
  END;



(* Initializes the lighting system. *)
  PROCEDURE Init;
  BEGIN
    DarkColour := al_makecol (DARK_COMPONENT, DARK_COMPONENT, DARK_COMPONENT);
    LightMap := al_create_bitmap (640, 480);
    al_clear_bitmap (LightMap);
    BigLight := CreateSpotlight (400);
  END;



(* Sets the sprite index to track by the big light. *)
  PROCEDURE SetBigLightSpriteIndex (SprIndex: INTEGER);
  BEGIN
    BigLightSpr := SprIndex;
  END;



(* Apply the lightmap. *)
  PROCEDURE Apply (Bmp: AL_BITMAPptr; ScrollX, ScrollY: INTEGER);
  BEGIN
  { Alpha channel. }
    al_set_write_alpha_blender;
    al_drawing_mode (AL_DRAW_MODE_TRANS, NIL, 0, 0);
    al_rectfill (LightMap, 0, 0, 640, 480, DarkColour);
    al_draw_trans_sprite (
      LightMap, BigLight,
      SpriteList[BigLightSpr].X - ScrollX - 168,
      SpriteList[BigLightSpr].Y - ScrollY - 200);
  { Set alpha blender and apply the lightmap. }
    al_drawing_mode (AL_DRAW_MODE_SOLID, NIL, 0, 0);
    al_set_alpha_blender;
    al_draw_trans_sprite (Bmp, LightMap, 0, 0);
  { Restore a not-alpha blender. }
    al_set_trans_blender (0, 0, 0, 128);
  END;



(* Releases resources. *)
  PROCEDURE ShutDown;
  BEGIN
    al_destroy_bitmap (LightMap);
    al_destroy_bitmap (BigLight);
  END;

END.
