UNIT mapmaker;
(* Duke of Dragonfear: Deathland.
 *	An action game for the 1st Pascal Game Development Mini Contest.
 *
 * Creates the maps for the game.
 *)
(*
  Copyright (c) 2011 Guillermo Martínez J.

  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software
    in a product, an acknowledgment in the product documentation would be
    appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
    distribution.
*)

INTERFACE

(* Creates the map.  It initializes sprites and tilemap. *)
  PROCEDURE CreateLevel;

IMPLEMENTATION

USES
  Enemies, Player, Sprites, TileMap;

CONST
(* Room type. *)
  WALKABLE = ORD (' ');
  NO_WALKABLE = ORD ('X');

  ZOMBIE_LAND = ORD ('z');
  SKELETON_PIT = ORD ('s');

  PLAYER_PLACE = ORD ('p');

(* Size of each room in tiles. *)
  TILES_ROOM = 8;

VAR
(* The map is a grid of rooms. *)
  Map: ARRAY [1..100] OF INTEGER;



(* Returns the upper left (north-east) tile of the map position. *)
  PROCEDURE FirstTileOfPosition (CONST Position: INTEGER; OUT oX, oY: INTEGER);
  BEGIN
  { I'm not sure why it's (TILES_ROOM - 2) instead of (TILES_ROOM - 1) but it
    works.

    Also note that in 1..100 interval "X" starts at 1 but "Y" starts at 0! }
    oX := (((Position MOD 10) - 1) * TILES_ROOM) - (TILES_ROOM - 2);
    oY := (((Position DIV 10)    ) * TILES_ROOM) - (TILES_ROOM - 2);
  END;



(* Adds enemies. *)
  PROCEDURE AddEnemies (Position: INTEGER);
  VAR
    Ix, Iy, NumEnemies: INTEGER;
  BEGIN
    FirstTileOfPosition (Position, Ix, Iy);
    CASE Map[Position] OF
    ZOMBIE_LAND:
      FOR NumEnemies := 1 TO 5 DO
	EnemyList.Add (TZombie.Create (NextFreeSprite, Ix + Random (TILES_ROOM), Iy + Random (TILES_ROOM)));
    SKELETON_PIT:
      FOR NumEnemies := 1 TO 3 DO
	EnemyList.Add (TSkeleton.Create (NextFreeSprite, Ix + Random (TILES_ROOM), Iy + Random (TILES_ROOM)));
    END;
  END;



(* Set the player position. *)
  PROCEDURE SetPlayerPosition (Position: INTEGER);
  VAR
    Ix, Iy: INTEGER;
  BEGIN
    FirstTileOfPosition (Position, Ix, Iy);
    (PlayerList.Actor[0] AS TPlayerActor).Init (Ix + 3, Iy + 4);
    (PlayerList.Actor[1] AS TPlayerActor).Init (Ix + 4, Iy + 4);
  END;



(* Helper procedure to add a block. *)
  PROCEDURE AddWallRoom (Position: INTEGER);
  VAR
    Ix, Iy, X, Y: INTEGER;
  BEGIN
    FirstTileOfPosition (Position, Ix, Iy);
  { If room at south is walkable then it doesn't fill the "southest" tiles to
    reserver room for the wall. }
    IF Map[Position + 10] <> NO_WALKABLE THEN
    BEGIN
      FOR Y := 0 TO (TILES_ROOM - 4) DO
	FOR X := 0 TO (TILES_ROOM - 1) DO
	  Board[Ix + X, Iy + Y] := 5;
    END
    ELSE
      FOR Y := 0 TO (TILES_ROOM - 1) DO
	FOR X := 0 TO (TILES_ROOM - 1) DO
	  Board[Ix + X, Iy + Y] := 5;
  END;



(* Fixes the map, putting edges and so. *)
  PROCEDURE FixMap;
  VAR
    Tw, Th, Val, X, Y: INTEGER;

  (* Helper functions to know if tile is border. *)
    FUNCTION Left: BOOLEAN;
    BEGIN
      Left := (X = 1) OR (tilemap.Board[X-1, Y] > 4);
    END;

    FUNCTION Right: BOOLEAN;
    BEGIN
      Right := (X = tilemap.Width) OR (tilemap.Board[X+1, Y] > 4);
    END;

    FUNCTION Up: BOOLEAN;
    BEGIN
      Up := (Y = 1) OR (tilemap.Board[X, Y-1] > 4);
    END;

    FUNCTION Down: BOOLEAN;
    BEGIN
      Down := (Y = tilemap.Height) OR (tilemap.Board[X, Y+1] > 4);
    END;

  BEGIN
  { Look for top and bottom. }
    Tw := 0; Th := 0;
    FOR Y := 1 TO tilemap.Height DO
      FOR X := 1 TO tilemap.Width DO
      BEGIN
	IF (tilemap.Board[X, Y] = 5) THEN
	BEGIN
	  IF NOT Up THEN
	  BEGIN
	    tilemap.Board[X, Y] := 58 + Tw;
	    Tw := 1 - Tw;
	  END
	  ELSE IF NOT Down THEN
	  BEGIN
	    FOR Val := 0 TO 2 DO
	      tilemap.Board[X, Y + Val] := 10 + Tw + (Val * 8);
	    Tw := 1 - Tw;
	  END;
	END;
      END;
  { Look for left and right. }
    FOR Y := 1 TO tilemap.Height DO
    BEGIN
      FOR X := 1 TO tilemap.Width DO
      BEGIN
	IF (tilemap.Board[X, Y] = 5) THEN
	BEGIN
	  IF NOT Right THEN
	  BEGIN
	    IF Th = 0 THEN
	      tilemap.Board[X, Y] := 32
	    ELSE
	      tilemap.Board[X, Y] := 40;
	  END
	  ELSE IF NOT Left THEN
	  BEGIN
	    IF Th = 0 THEN
	      tilemap.Board[X, Y] := 37
	    ELSE
	      tilemap.Board[X, Y] := 45;
	  END;
	END;
      END;
      Th := 1 - Th;
    END;
  { Look for corners. }
    FOR Y := 1 TO tilemap.Height DO
      FOR X := 1 TO tilemap.Width DO
      BEGIN
	IF (tilemap.Board[X, Y] = 10) OR (tilemap.Board[X, Y] = 11) THEN
	BEGIN
	  IF tilemap.Board[X-1, Y] = 5 THEN
	  BEGIN
	    FOR Val := 0 TO 2 DO
	    BEGIN
	      tilemap.Board[X-1, Y + Val] := 8 + (Val * 8);
	      tilemap.Board[X,   Y + Val] := 9 + (Val * 8);
	    END;
	    tilemap.Board[X,   Y + 3] := 33;
	  END
	  ELSE IF tilemap.Board[X-1, Y] < 5 THEN
	  BEGIN
	    FOR Val := 0 TO 3 DO
	    BEGIN
	      tilemap.Board[X-1, Y + Val] := 38 + (Val * 8);
	      tilemap.Board[X,   Y + Val] := 39 + (Val * 8);
	    END;
	  END
	  ELSE IF tilemap.Board[X+1, Y] < 5 THEN
	  BEGIN
	    FOR Val := 0 TO 3 DO
	    BEGIN
	      tilemap.Board[X+1, Y + Val] := 7 + (Val * 8);
	      tilemap.Board[X,   Y + Val] := 6 + (Val * 8);
	    END;
	  END
	  ELSE IF tilemap.Board[X+1, Y] = 5 THEN
	  BEGIN
	    FOR Val := 0 TO 2 DO
	    BEGIN
	      tilemap.Board[X+1, Y + Val] := 13 + (Val * 8);
	      tilemap.Board[X,   Y + Val] := 12 + (Val * 8);
	    END;
	  END;
	END
	ELSE IF (tilemap.Board[X, Y] = 58) OR (tilemap.Board[X, Y] = 59) THEN
	BEGIN
	  IF tilemap.Board[X-1, Y] = 5 THEN
	  BEGIN
	    tilemap.Board[X-1, Y] := 56;
	    tilemap.Board[X, Y] := 57;
	    tilemap.Board[X-1, Y-1] := 48;
	    tilemap.Board[X, Y-1] := 49;
	  END
	  ELSE IF tilemap.Board[X-1, Y] < 5 THEN
	  BEGIN
	    tilemap.Board[X-1, Y] := 42;
	    tilemap.Board[X, Y] := 43;
	    tilemap.Board[X-1, Y-1] := 34;
	    tilemap.Board[X, Y-1] := 35;
	  END
	  ELSE IF tilemap.Board[X+1, Y] < 5 THEN
	  BEGIN
	    tilemap.Board[X+1, Y] := 51;
	    tilemap.Board[X, Y] := 50;
	    tilemap.Board[X, Y-1] := 41;
	  END
	  ELSE IF tilemap.Board[X+1, Y] = 5 THEN
	  BEGIN
	    tilemap.Board[X+1, Y] := 61;
	    tilemap.Board[X, Y] := 60;
	    tilemap.Board[X+1, Y-1] := 53;
	    tilemap.Board[X, Y-1] := 52;
	  END;
	END;
      END;
  END;



(* Creates the map.  It initializes sprites and tilemap.

   First it creates a 10x10 map like this:
   ( X = wall, e = enemies, i = items ... )

  XXXXXXXXXX
  X    X   X
  X i e    X
  X  X X X X
  X e    i X
  X X  X   X
  X  X e   X
  X e iX X X
  X  X  e  X
  XXXXXXXXXX

  Then it uses this grid to fill the tilemap with apropriate tiles and
  creates the enemies, items, etc.
*)
  PROCEDURE CreateLevel;
  VAR
    Cnt, Ndx, X, Y: INTEGER;
  BEGIN
  { Puts the border walls and cleans the inside. }
    FOR Cnt := 1 TO 100 DO
    BEGIN
      IF (Cnt < 12) OR (Cnt > 90)
      OR (10 * (Cnt DIV 10) = Cnt) OR (10 * (Cnt DIV 10) = (Cnt -1)) THEN
	Map[Cnt] := NO_WALKABLE
      ELSE
	Map[Cnt] := WALKABLE;
    END;
  { Starting place. }
    Map[55] := PLAYER_PLACE;
  { Puts room types randomly. }
  { Enemies. }
    Cnt := 1;
    REPEAT
      Ndx := Random (100) + 1;
      IF Map[Ndx] = WALKABLE THEN
      BEGIN
	IF Cnt MOD 2 = 1 THEN
	  Map[Ndx] := ZOMBIE_LAND
	ELSE
	  Map[Ndx] := SKELETON_PIT;
	INC (Cnt);
      END;
    UNTIL Cnt > 10;
  { Non-walkable rooms. }
    Cnt := 1;
    REPEAT
      Ndx := Random (100) + 1;
      IF Map[Ndx] = WALKABLE THEN
      BEGIN
	Map[Ndx] := NO_WALKABLE;
	INC (Cnt);
      END;
    UNTIL Cnt > 10;
(* For DEBUG, it prints the current map in the console. *)
    FOR Cnt := 1 TO 100 DO
    BEGIN
      IF (Cnt MOD 10) = 1 THEN
	WriteLn;
      Write (CHR (Map[Cnt]));
    END;
    WriteLn;
  { Initializes the tilemap. }
    FOR Y := 1 TO tilemap.Width DO
      FOR X := 1 TO tilemap.Height DO
      BEGIN
	IF (X = 1) OR (Y = 1)
	OR (X = tilemap.Width) OR (Y = tilemap.Height) THEN
	{ Border is wall. }
	  tilemap.Board[X, Y] := 5
	ELSE
	{ Puts random floor. }
	  tilemap.Board[X, Y] := random (5);
      END;
  { Now, sets walkable and non-walkable tiles in the tilemap for each room,
    and sets sprites. }
    Sprites.Init;
    FOR Cnt := 1 TO 100 DO
    BEGIN
      IF (12 > Cnt) OR (Cnt > 90)
      OR (10 * (Cnt DIV 10) = Cnt) OR (10 * (Cnt DIV 10) = Cnt - 1)
      THEN
	CONTINUE;
      CASE Map[Cnt] OF
	NO_WALKABLE:
	  AddWallRoom (Cnt);
	PLAYER_PLACE:
	  SetPlayerPosition (Cnt);
	ZOMBIE_LAND, SKELETON_PIT:
	  AddEnemies (Cnt);
      END;
    END;
  { Finally, uses next procedure to beautify the tilemap, finding edeges,
    corners, etc. }
    FixMap;
  END;

END.
