UNIT player;
(* Duke of Dragonfear: Deathland.
 *	An action game for the 1st Pascal Game Development Mini Contest.
 *
 * The player.
 *
 * Note that attack sprite collision checking is done at the enemies actor.
 *)
(*
  Copyright (c) 2011 Guillermo Martínez J.

  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software
    in a product, an acknowledgment in the product documentation would be
    appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
    distribution.
*)

INTERFACE

USES
  actor,
  allegro;

TYPE
(* Player characters have some stuf in common. *)
  TPlayerActor = CLASS (TCharacterActor)
  PRIVATE
  (* To keep track of health point. *)
    fOldHealth: INTEGER;
  (* The health-metter. *)
    fHealthMetterBitmap: AL_BITMAPptr;
  PUBLIC
  (* Creates the character. *)
    CONSTRUCTOR Create (aPictureList: TCharacterPictureList; aSampleList: TSoundList; aSprIndex: INTEGER); OVERRIDE;
  (* Destroys the object and releases resources. *)
    DESTRUCTOR Destroy; OVERRIDE;
  (* Initializes the character. *)
    PROCEDURE Init (Px, Py: INTEGER); VIRTUAL;
  (* To get the scroll position. *)
    PROCEDURE Scroll (OUT X, Y: INTEGER); INLINE;
  (* Draws the health-metter in the given bitmap. *)
    PROCEDURE DrawHealth (CONST Bmp: AL_BITMAPptr; CONST X, Y: INTEGER);
  END;



(* The Knight character. *)
  TKnight = CLASS (TPlayerActor)
  PRIVATE
  (* To prevent continuous attack. *)
    fDidAttack: BOOLEAN;
  PUBLIC
  (* Creates the player. *)
    CONSTRUCTOR Create; OVERLOAD;
  (* Initializes the player. *)
    PROCEDURE Init (Px, Py: INTEGER); OVERRIDE;
  (* Process player input. *)
    PROCEDURE Update; OVERRIDE;
  (* *)
    PROCEDURE Attacked (Strength: INTEGER); OVERRIDE;
  END;



(* He carry a torch that lights the scene, and walks beside the player. *)
  TTorchMan = CLASS (TPlayerActor)
  PRIVATE
  (* The character to follow. *)
    fMainCharacter: TActor;
  PUBLIC
  (* Creates the torch-man. *)
    CONSTRUCTOR Create; OVERLOAD;
  (* Initializes the player. *)
    PROCEDURE Init (Px, Py: INTEGER); OVERRIDE;
  (* Process player input. *)
    PROCEDURE Update; OVERRIDE;
  (* *)
    PROCEDURE Attacked (Strength: INTEGER); OVERRIDE;
  (* The character to follow. *)
    PROPERTY MainCharacter: TActor READ fMainCharacter WRITE fMainCharacter;
  END;

CONST
  PLAYER_STRENGTH = 1;

VAR
(* The enemies needs to know where the player is! *)
  PlayerList: TActorList;

IMPLEMENTATION

USES
  gamedata, Geometry, graphsys, Lighting, Sprites, TileMap, usrInput;



(****************
 * TPlayerActor *
 ****************)

(* Creates the player. *)
  CONSTRUCTOR TPlayerActor.Create (aPictureList: TCharacterPictureList; aSampleList: TSoundList; aSprIndex: INTEGER);
  BEGIN
    INHERITED Create (aPictureList, aSampleList, aSprIndex);
    fHealthMetterBitmap := al_create_bitmap (HP_FullBitmap^.w, HP_FullBitmap^.h);
    fTypeId := atPlayer;
  END;



(* Destroys the object and releases resources. *)
  DESTRUCTOR TPlayerActor.Destroy;
  BEGIN
    al_destroy_bitmap (fHealthMetterBitmap);
    INHERITED Destroy;
  END;



(* Initializes the player. *)
  PROCEDURE TPlayerActor.Init (Px, Py: INTEGER);
  BEGIN
    fHealthPoints := 10;
    fOldHealth := 10;
    fSprIndex := Sprites.NextFreeSprite;
    fStatus := aaStop;
    fDirection := RS_S;
  { Copy the full HP-meter to the current HP-meter. }
    al_blit (HP_FullBitmap, fHealthMetterBitmap, 0, 0, 0, 0, fHealthMetterBitmap^.w, fHealthMetterBitmap^.h);
    WITH SpriteList[fSprIndex] DO
    BEGIN
      X := Px * TSIZE;
      Y := Py * TSIZE;
      OffX :=  -95;
      OffY := -127;
      Picture := fPictures.Picture[RS_S, RS_STAND];
    END;
  END;



(* To get the scroll position. *)
  PROCEDURE TPlayerActor.Scroll (OUT X, Y: INTEGER);
  BEGIN
    X := SpriteList[fSprIndex].X - (640 DIV 2) + (TSIZE DIV 2);
    Y := SpriteList[fSprIndex].Y - (480 DIV 2) + (TSIZE DIV 2);
    TileMap.FixScroll (ActivePage, X, Y, X, Y);
  END;



(* Draws the health-metter in the given bitmap. *)
  PROCEDURE TPlayerActor.DrawHealth (CONST Bmp: AL_BITMAPptr; CONST X, Y: INTEGER);
  CONST
  (* Health-metter sizes. *)
    HM_WIDTH  = 31;
    HM_HEIGHT = 27;
    HM_INNER_HEIGHT = 24;
  VAR
    PosY: INTEGER;
  BEGIN
    IF fOldHealth <> fHealthPoints THEN
    BEGIN
    { Copy the bottom part of the full health-metter over the empty one. }
      PosY := HM_INNER_HEIGHT - ((HM_INNER_HEIGHT * fHealthPoints) DIV 10) + 1;
      al_blit (HP_EmptyBitmap, fHealthMetterBitmap, 0,    0, 0,    0, HM_WIDTH, HM_HEIGHT);
      al_blit (HP_FullBitmap,  fHealthMetterBitmap, 0, PosY, 0, PosY, HM_WIDTH, HM_HEIGHT);
      fOldHealth := fHealthPoints;
    END;
    al_draw_Sprite (Bmp, fHealthMetterBitmap, X, Y);
  END;



(***********
 * TKnight *
 ***********)

(* Creates the player. *)
  CONSTRUCTOR TKnight.Create;
  BEGIN
    INHERITED Create (gamedata.PlayerPictures, gamedata.PlayerSamples, Sprites.NextFreeSprite);
    fDidAttack := FALSE;
  END;



(* Initializes the player. *)
  PROCEDURE TKnight.Init (Px, Py: INTEGER);
  BEGIN
    INHERITED Init (Px, Py);
    fWeaponSprIndex := Sprites.NextFreeSprite;
  END;



(* Process player input. *)
  PROCEDURE TKnight.Update;

  (* Helper function to know if it starts ataking. *)
    FUNCTION CheckAttack: BOOLEAN;
    BEGIN
      CheckAttack := FALSE;
      IF NOT fDidAttack THEN
      BEGIN
	IF UserInput[1].Button (BTN_ATTACK) THEN
	BEGIN
	  InitAttack;
	  CheckAttack := TRUE;
	  fDidAttack := TRUE;
	END;
      END
      ELSE
	fDidAttack := UserInput[1].Button (BTN_ATTACK);
    END;

  VAR
    Dir: INTEGER;
  BEGIN
    Dir := UserInput[1].LeverDir;
  { Each status has different process. }
    CASE fStatus OF
      aaAching: BEGIN
	INC (fCounter);
	IF fCounter = 20 THEN
	BEGIN
	  IF fHealthPoints > 0 THEN
	  BEGIN
	    fStatus := aaStop;
	    SpriteList[fSprIndex].Picture :=
		fPictures.Picture[fDirection, (fCounter DIV 10) + 1];
	  END
	  ELSE BEGIN
	    SpriteList[fSprIndex].Picture := fPictures.Picture[fDirection, RS_DEAD];
	    fStatus := aaDead;
	  END;
	END;
      END;
      aaAttacking: BEGIN
	INC (fCounter);
	IF fCounter = 10 THEN
	  SpriteList[fWeaponSprIndex].Y := SPR_Y_OUT
	ELSE
	IF fCounter = 25 THEN
	BEGIN
	  fCounter := 0;
	  SpriteList[fSprIndex].Picture := fPictures.Picture[fDirection, RS_STAND];
	  fStatus := aaStop;
	END;
      END;
      aaWalking:
	IF NOT CheckAttack THEN
	BEGIN
	  IF Dir <> DIR_NOP THEN
	  BEGIN
	  { Checks walking direction. }
	    fDirection := Dir;
	    IF (Dir IN DIR_G_LEFT) AND NOT CheckLeftCollision (fSprIndex) THEN
	      DEC (SpriteList[fSprIndex].x);
	    IF (Dir IN DIR_G_RIGHT) AND NOT CheckRightCollision (fSprIndex) THEN
	      INC (SpriteList[fSprIndex].x);
	    IF (Dir IN DIR_G_UP) AND NOT CheckCeilCollision (fSprIndex) THEN
	      DEC (SpriteList[fSprIndex].y);
	    IF (Dir IN DIR_G_DOWN) AND NOT CheckGroundCollision (fSprIndex) THEN
	      INC (SpriteList[fSprIndex].y);
	  { Checks if the sprite should be changed. }
	    INC (fCounter);
	    IF fCounter >= 40 THEN
	      fCounter := 0;
	    SpriteList[fSprIndex].Picture := fPictures.Picture[fDirection, (fCounter DIV 10) + 1];
	  END
	  ELSE
	    fStatus := aaStop;
	END;
      aaDead: BEGIN
	{ Does nothing (at the moment). }
      END;
      ELSE BEGIN
      { Check if starts moving. }
	IF Dir <> DIR_NOP THEN
	BEGIN
	  fStatus := aaWalking;
	  IF fDirection <> Dir THEN
	  BEGIN
	    fDirection := Dir;
	    fCounter := 0;
	    SpriteList[fSprIndex].Picture := fPictures.Picture[fDirection, RS_WALK_BEGIN];
	  END;
	END;
      { Check if starts an attack. }
	CheckAttack;
      END;
    END;
  END;



(* *)
  PROCEDURE TKnight.Attacked (Strength: INTEGER);
  BEGIN
    IF fStatus = aaAttacking THEN
    BEGIN
      Sprites.PlaySoundSample (fSprIndex, fSamples.Dodge);
      SpriteList[fSprIndex].Picture := fPictures.Picture[fDirection, RS_DEFENSE];
    END
    ELSE IF fStatus IN [aaWalking, aaStop] THEN
      InitHurt (Strength);
  END;



(*************
 * TTorchMan *
 *************)

CONST
  MIN_FOLLOW_DISTANCE = TSIZE;
  MIN_FOLLOW_DISTANCE2 = MIN_FOLLOW_DISTANCE * MIN_FOLLOW_DISTANCE;
  MAX_FOLLOW_DISTANCE = 180;
  MAX_FOLLOW_DISTANCE2 = MAX_FOLLOW_DISTANCE * MAX_FOLLOW_DISTANCE;

  WAITING_COUNTER = 100;



(* Creates the torch-man. *)
  CONSTRUCTOR TTorchMan.Create;
  BEGIN
    INHERITED Create (gamedata.TorchManPictures, gamedata.PlayerSamples, Sprites.NextFreeSprite);
  END;



(* Initializes the torch-man. *)
  PROCEDURE TTorchMan.Init (Px, Py: INTEGER);
  BEGIN
    INHERITED Init (Px, Py);
    Lighting.SetBigLightSpriteIndex (SELF.SpriteIndex);
  END;



(* Process player input. *)
  PROCEDURE TTorchMan.Update;
  BEGIN
  { Each status has different process. }
    CASE fStatus OF
      aaAching: BEGIN
	INC (fCounter);
	IF fCounter = 20 THEN
	BEGIN
	  IF fHealthPoints > 0 THEN
	  BEGIN
	    fStatus := aaStop;
	    SpriteList[fSprIndex].Picture :=
		fPictures.Picture[fDirection, (fCounter DIV 10) + 1];
	  END
	  ELSE BEGIN
	    SpriteList[fSprIndex].Picture := fPictures.Picture[fDirection, RS_DEAD];
	    fStatus := aaDead;
	  END;
	END;
      END;
      aaWaiting: BEGIN
      { Waits until MainCharacter isn't attacking or aching. }
	IF ((fMainCharacter AS TCharacterActor).Status IN [aaAching, aaAttacking])
	THEN
	  fCounter := WAITING_COUNTER
	ELSE BEGIN
	  DEC (fCounter);
	  IF fCounter < 1 THEN
	  BEGIN
	    fStatus := aaStop;
	    SpriteList[fSprIndex].Picture := fPictures.Picture[fDirection, RS_STAND];
	  END;
	END;
      END;
      aaWalking: BEGIN
      { Checks walking direction. }
	IF (fDirection IN DIR_G_LEFT) AND NOT CheckLeftCollision (fSprIndex) THEN
	  DEC (SpriteList[fSprIndex].x);
	IF (fDirection IN DIR_G_RIGHT) AND NOT CheckRightCollision (fSprIndex) THEN
	  INC (SpriteList[fSprIndex].x);
	IF (fDirection IN DIR_G_UP) AND NOT CheckCeilCollision (fSprIndex) THEN
	  DEC (SpriteList[fSprIndex].y);
	IF (fDirection IN DIR_G_DOWN) AND NOT CheckGroundCollision (fSprIndex) THEN
	  INC (SpriteList[fSprIndex].y);
      { Checks if the sprite should be changed. }
	INC (fCounter);
	IF fCounter >= 40 THEN
	BEGIN
	  fDirection := LookAt (
	    SpriteList[fSprIndex].X, SpriteList[fSprIndex].Y,
	    SpriteList[fMainCharacter.SpriteIndex].X, SpriteList[fMainCharacter.SpriteIndex].Y
	  );
	  fCounter := 0;
	  IF ((fMainCharacter AS TCharacterActor).Status IN [aaAching, aaAttacking])
	  OR (Distance2 (
	    SpriteList[fMainCharacter.SpriteIndex].X, SpriteList[fMainCharacter.SpriteIndex].Y,
	    SpriteList[fSprIndex].X, SpriteList[fSprIndex].Y) <= MIN_FOLLOW_DISTANCE2)
	  THEN
	    fStatus := aaStop;
	END
	ELSE IF Distance2 (
	  SpriteList[fMainCharacter.SpriteIndex].X, SpriteList[fMainCharacter.SpriteIndex].Y,
	  SpriteList[fSprIndex].X, SpriteList[fSprIndex].Y) <= MIN_FOLLOW_DISTANCE2
	THEN
	  fStatus := aaStop;
	SpriteList[fSprIndex].Picture := fPictures.Picture[fDirection, (fCounter DIV 10) + 1];
      END;
      aaDead: BEGIN
	{ Does nothing (at the moment). }
      END;
      ELSE BEGIN
      { Is player is doing something, it waits. }
	IF (fMainCharacter AS TCharacterActor).Status IN [aaAching, aaAttacking] THEN
	BEGIN
	  fCounter := WAITING_COUNTER;
	  fStatus := aaWaiting;
	END
	ELSE BEGIN
	{ Try to keep near the player. }
	  fDirection := LookAt (
	    SpriteList[fSprIndex].X, SpriteList[fSprIndex].Y,
	    SpriteList[fMainCharacter.SpriteIndex].X, SpriteList[fMainCharacter.SpriteIndex].Y
	  );
	  IF (fMainCharacter AS TCharacterActor).Status IN [aaWaiting, aaScentry, aaStop] THEN
	  BEGIN
	    IF Distance2 (
	      SpriteList[fMainCharacter.SpriteIndex].X, SpriteList[fMainCharacter.SpriteIndex].Y,
	      SpriteList[fSprIndex].X, SpriteList[fSprIndex].Y) > MIN_FOLLOW_DISTANCE2
	    THEN BEGIN
	      fStatus := aaWalking;
	    END;
	  END
	  ELSE IF Distance2 (
	    SpriteList[fMainCharacter.SpriteIndex].X, SpriteList[fMainCharacter.SpriteIndex].Y,
	    SpriteList[fSprIndex].X, SpriteList[fSprIndex].Y) >= MAX_FOLLOW_DISTANCE2
	  THEN BEGIN
	    fCounter := 0;
	    fStatus := aaWalking;
	  END;
	END;
	SpriteList[fSprIndex].Picture := fPictures.Picture[fDirection, RS_STAND];
      END;
    END;
  END;



(* *)
  PROCEDURE TTorchMan.Attacked (Strength: INTEGER);
  BEGIN
    IF fStatus <> aaDead THEN
    BEGIN
      IF (fStatus IN [aaWalking, aaStop]) OR (Random (10) > 1) THEN
	InitHurt (Strength)
      ELSE
      BEGIN
	Sprites.PlaySoundSample (fSprIndex, fSamples.Dodge);
	SpriteList[fSprIndex].Picture := fPictures.Picture[fDirection, RS_DEFENSE];
	fCounter := 40;
	fStatus := aaWaiting;
      END;
    END;
  END;

END.
