UNIT Sprites;
(* Program: Duke of Dragonfear: Deathland.
 * Description: Manages a list of sprites and draws them.
 *
 * It orders the sprites by its Y coordinate, to simulate 3D.
 *
 * Tilemap collision are calculated as if sprite hass same size as tiles.
 * That's why it needs the offset to draw them.
 *
 * Author: Ñuño Martínez <niunio@users.sourceforge.net>
 *)
(*
  Copyright (c) 2011 Guillermo Martínez J.

  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software
    in a product, an acknowledgment in the product documentation would be
    appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
    distribution.
*)

INTERFACE

USES
  allegro;

TYPE
(* Pointer to sprite. *)
  TSPRITEptr = ^TSPRITE;
(* Sprite information. *)
  TSPRITE = RECORD
  { Sprite world coordinates. }
    x, y: INTEGER;
  { Offset from world coordinate to bitmap upper-left pixel. }
    OffX, OffY: INTEGER;
  { Reference to the RLE picture. }
    Picture: AL_RLE_SPRITEptr;
  END;

CONST
{ Number of sprites. }
  NUM_SPRITES = 250;
{ This value was used by an old graphic chipset to disable hardware sprite
  planes.  It's used as an homenage. }
  SPR_Y_OUT = -129;
{ Sprite size.  That's not the bitmap size, but the bounding box of the sprite. }
  SPR_SIZE = 64;

VAR
(* Array with sprite information. *)
  SpriteList: ARRAY [0..(NUM_SPRITES - 1)] OF TSPRITE;



(* Initializes all sprites.  That is moves them out the board and sets them as
   "inactive".  It clears Offsets also *)
  PROCEDURE Init;

(* Returns the next free sprite index.  Useful to dynamicly add sprites.
   Returns -1 if there's no space for new sprites.  To make it works sprites
   should be "deactivated" by assigning SPR_Y_OUT to its Y coordinate. *)
  FUNCTION NextFreeSprite: INTEGER;

(* Checks if sprites collided. *)
  FUNCTION Collision (CONST Spr1, Spr2: INTEGER): BOOLEAN; INLINE;

(* Checks if there's a ground tile below the sprite. *)
  FUNCTION CheckGroundCollision (CONST SprNdx: INTEGER): BOOLEAN;

(* Checks if there's a ceil tile above the sprite. *)
  FUNCTION CheckCeilCollision (CONST SprNdx: INTEGER): BOOLEAN;

(* Checks if there's a ground tile on the left of the sprite. *)
  FUNCTION CheckLeftCollision (CONST SprNdx: INTEGER): BOOLEAN;

(* Checks if there's a ground tile on the right of the sprite. *)
  FUNCTION CheckRightCollision (CONST SprNdx: INTEGER): BOOLEAN;

(* Checks if the given tile collides with the sprite.  Returns a number
   to identify the quadrants witch the tile is as a combination (OR) of
   the following values:
   0 - No collision at all.
   1 - Top-left quadrant.
   2 - Top-right quadrant.
   4 - Bottom-left quadrant.
   8 - Bottom-Right quadrant.
   As example, if the tile is in both top quadrants, it will return:
    (1 OR 2) = 3 *)
  FUNCTION CheckCollisionWith (CONST SprNdx, TileVal: INTEGER): INTEGER;

(* Draws the sprites in the given bitmap at the given scroll coordinates. *)
  PROCEDURE Draw (Bmp: AL_BITMAPptr; ScrollX, ScrollY: INTEGER);

(* Draws the sprites using the lightmap from unit "Lighting" to light them. *)
  PROCEDURE DrawLighted (Bmp: AL_BITMAPptr; ScrollX, ScrollY: INTEGER);

(* Plays a sound sample changing the stereo 'pan' as the X coordinate
   of the sprite. *)
  PROCEDURE PlaySoundSample (Spr: INTEGER; Sample: AL_SAMPLEptr); INLINE;



IMPLEMENTATION

USES
  gamedata, { To acces to de game datafile. }
  Lighting,
  tilemap;  { Tilemap management. }

VAR
(* Current free sprite. *)
  FreeSprite: INTEGER;
(* Drawing order dictionary.  The sprite '1' is the 'top' sprite, the
  '2' is drawn below the '1', the '3' is drawn below '2' and '1' and so... *)
  SpriteOrder: ARRAY [1..NUM_SPRITES] OF INTEGER;
(* Needed for sound effects. *)
  LastScrollX: INTEGER;



(* Starts all sprites.  That is moves them out the board and sets them as
   "inactive". *)
  PROCEDURE Init;
  VAR
    Ndx: INTEGER;
  BEGIN
    FOR Ndx := 0 TO (NUM_SPRITES -1) DO
    BEGIN
      SpriteList[Ndx].y := SPR_Y_OUT;
      SpriteList[Ndx].OffX := 0;
      SpriteList[Ndx].OffY := 0;
      SpriteList[Ndx].Picture := NIL;
      SpriteOrder[Ndx + 1] := Ndx;
    END;
    FreeSprite := 0;
  END;



(* Returns the next free sprite index.  Useful to dynamicly add sprites.
   Returns -1 if there's no space for new sprites.  To make it works sprites
   should be "deactivated" by assigning SPR_Y_OUT to its Y coordinate. *)
  FUNCTION NextFreeSprite: INTEGER;
  BEGIN
    IF FreeSprite >= NUM_SPRITES THEN
    BEGIN
      FreeSprite := 0;
      WHILE (FreeSprite < NUM_SPRITES)
      AND (SpriteList[FreeSprite].Y <> SPR_Y_OUT) DO
	INC (FreeSprite);
      IF FreeSprite >= NUM_SPRITES THEN
      BEGIN
	FreeSprite := NUM_SPRITES;
	NextFreeSprite := -1;
      END;
    END
    ELSE BEGIN
    {TODO: What if it was yes assigned? }
      NextFreeSprite := FreeSprite;
      INC (FreeSprite);
    END;
  END;



(* Checks if sprites collided between them. *)
  FUNCTION Collision (CONST Spr1, Spr2: INTEGER): BOOLEAN;
  BEGIN
    Collision := (ABS (SpriteList[Spr1].X - SpriteList[Spr2].X) < SPR_SIZE)
	     AND (ABS (SpriteList[spr1].Y - SpriteList[spr2].Y) < SPR_SIZE);
  END;



(* Check if there's a ground tile below the sprite. *)
  FUNCTION CheckGroundCollision (CONST SprNdx: INTEGER): BOOLEAN;
  BEGIN
    CheckGroundCollision := FALSE;
  { First, check if the sprite is just above a tile. }
    IF SpriteList[SprNdx].y MOD TSIZE <> 0 THEN
      EXIT;
  { Check if the tile to check is not out of the map. }
  { Map should be surrounded by wall, so it's no necessary.
    IF SpriteList[SprNdx].y < (Tilemap.Height - 1) * TSiZE THEN
      CheckGroundCollision := TRUE
    ELSE
  }
  { Check the tile just below the sprite. }
    IF Tilemap.IsWall ((SpriteList[SprNdx].x DIV TSIZE) + 1,
	(SpriteList[SprNdx].y DIV TSIZE) + 2)
    THEN
      CheckGroundCollision := TRUE
    ELSE
    { Check if the sprite is in the middle of two tiles. }
      IF SpriteList[SprNdx].x MOD TSIZE <> 0 THEN
      { It is, so check again with the tile at side. }
	IF Tilemap.IsWall ((SpriteList[SprNdx].x DIV TSIZE) + 2,
		(SpriteList[SprNdx].y DIV TSIZE) + 2)
	THEN
	  CheckGroundCollision := TRUE;
  END;



(* Check if there's a ceil tile above the sprite. *)
  FUNCTION CheckCeilCollision (CONST SprNdx: INTEGER): BOOLEAN;
  BEGIN
    CheckCeilCollision := FALSE;
  { First, check if the sprite is just below a tile. }
    IF SpriteList[SprNdx].y MOD TSIZE <> 0 THEN
      EXIT;
  { Check if the tile is not out of the map. }
  { Map should be surrounded by wall, so it's no necessary.
    IF SpriteList[SprNdx].y DIV TSIZE < 1 THEN
      CheckCeilCollision := TRUE
    ELSE
  }
  { Check the tile just above the sprite. }
    IF Tilemap.IsWall ((SpriteList[SprNdx].x DIV TSIZE) + 1,
	(SpriteList[SprNdx].y DIV TSIZE))
    THEN
      CheckCeilCollision := TRUE
    ELSE
    { Check if the sprite is in the middle of two tiles. }
      IF SpriteList[SprNdx].x MOD TSIZE <> 0 THEN
      { It is, so check again with the other tile. }
	IF Tilemap.IsWall ((SpriteList[SprNdx].x DIV TSIZE) + 2,
		(SpriteList[SprNdx].y DIV TSIZE))
	THEN
	  CheckCeilCollision := TRUE;
    END;



(* Check if there's a ground tile on the left of the sprite. *)
  FUNCTION CheckLeftCollision (CONST SprNdx: INTEGER): BOOLEAN;
  BEGIN
    CheckLeftCollision := FALSE;
  { First, check if the sprite is just in the right side of a tile. }
    IF SpriteList[SprNdx].x MOD TSIZE <> 0 THEN
      EXIT;
  { Now, Check if the tile to chek is inside the map. }
  { Map should be surrounded by wall, so it's no necessary.
    IF SpriteList[SprNdx].x DIV TSIZE < 1 THEN
    BEGIN
      CheckLeftCollision := TRUE;
      EXIT;
    END;
  }
  { Check the tile on the left of the sprite. }
    IF Tilemap.IsWall ((SpriteList[SprNdx].x DIV TSIZE),
	(SpriteList[SprNdx].y DIV TSIZE) + 1)
    THEN
      CheckLeftCollision := TRUE
    ELSE
    { Check if the sprite is in the middle of two tiles. }
      IF SpriteList[SprNdx].y MOD TSIZE <> 0 THEN
      { It is, so check again with the other tile. }
	IF Tilemap.IsWall ((SpriteList[SprNdx].x DIV TSIZE),
		(SpriteList[SprNdx].y DIV TSIZE) + 2)
	THEN
	  CheckLeftCollision := TRUE;
  END;



(* Check if there's a ground tile on the right of the sprite. *)
  FUNCTION CheckRightCollision (CONST SprNdx: INTEGER): BOOLEAN;
  BEGIN
    CheckRightCollision := FALSE;
  { First, check if the sprite is just in the right side of a tile. }
    IF SpriteList[SprNdx].x MOD TSIZE <> 0 THEN
      EXIT;
  { Now, Check if the tile to chek is inside the map. }
  { Map should be surrounded by wall, so it's no necessary.
    IF (SpriteList[SprNdx].x DIV TSIZE) + 2 > Tilemap.Width THEN
    BEGIN
      CheckRightCollision := TRUE;
      EXIT;
    END;
  }
  { Check the tile on the right of the sprite. }
    IF Tilemap.IsWall ((SpriteList[SprNdx].x DIV TSIZE) + 2,
	(SpriteList[SprNdx].y DIV TSIZE) + 1)
    THEN
      CheckRightCollision := TRUE
    ELSE
    { Check if the sprite is in the middle of two tiles. }
      IF SpriteList[SprNdx].y MOD TSIZE <> 0 THEN
      { It is, so check again with the other tile. }
	IF Tilemap.IsWall ((SpriteList[SprNdx].x DIV TSIZE) + 2,
		(SpriteList[SprNdx].y DIV TSIZE) + 2)
	THEN
	  CheckRightCollision := TRUE;
  END;



(* Checks if the given tile collides with the sprite.  Returns a number
   to identify the quadrants witch the tile is as a combination (OR) of
   the following values:
    0 - No collision at all.
    1 - Top-left quadrant.
    2 - Top-right quadrant.
    4 - Bottom-left quadrant.
    8 - Bottom-Right quadrant.
   As example, if the tile is in both top quadrants, it will return:
   (1 OR 2) = 3 *)
  FUNCTION CheckCollisionWith (CONST SprNdx, TileVal: INTEGER): INTEGER;
  VAR
    Tmp, Tx, Ty: INTEGER;
    BEGIN
    Tmp := 0;
  { Get the tile coordinates of the top left pixel of the sprite. }
    Tx := (SpriteList[SprNdx].x DIV TSIZE) + 1;
    Ty := (SpriteList[SprNdx].y DIV TSIZE) + 1;
  { Check only if where inside the map and the tile. }
    IF (0 < Tx) AND (Tx <= Tilemap.Width) AND (0 < Ty) AND (Ty <= Tilemap.Height) THEN
      IF Tilemap.Board[Tx, Ty] = TileVal THEN
	Tmp := 1;
    INC (Tx);
    IF (SpriteList[SprNdx].x MOD TSIZE <> 0)
    AND (0 < Tx) AND (Tx <= Tilemap.Width) AND (0 < Ty) AND (Ty <= Tilemap.Height) THEN
      IF Tilemap.Board[Tx, Ty] = TileVal THEN
	Tmp := Tmp OR 2;
    INC (Ty);
    IF (SpriteList[SprNdx].x MOD TSIZE <> 0)
    AND (SpriteList[SprNdx].y MOD TSIZE <> 0)
    AND (0 < Tx) AND (Tx <= Tilemap.Width) AND (0 < Ty) AND (Ty <= Tilemap.Height) THEN
      IF Tilemap.Board[Tx, Ty] = TileVal THEN
	Tmp := Tmp OR 8;
    DEC (Tx);
    IF (SpriteList[SprNdx].y MOD TSIZE <> 0)
    AND (0 < Tx) AND (Tx <= Tilemap.Width) AND (0 < Ty) AND (Ty <= Tilemap.Height) THEN
      IF Tilemap.Board[Tx, Ty] = TileVal THEN
	Tmp := Tmp OR 4;
  { Result. }
    CheckCollisionWith := Tmp;
  END;



(* Draws the sprites in the given bitmap at the given scroll coordinates. *)
  PROCEDURE Draw (Bmp: AL_BITMAPptr; ScrollX, ScrollY: INTEGER);
  VAR
    Ndx, Tmp: INTEGER;
    Posx, Posy: INTEGER;
    RLESprite: AL_RLE_SPRITEptr;
  BEGIN
  { Order sprites by Y.  Before you ask about the ordering algorithm, please
    read next paragraph:

    The "Quick-sort" is the fastest in the worst case (i.e. the list is fully
    unordered) but if the list is almost ordered (as the sprite list is) it is
    quite slow actually.  That's why I'm using a different one here. }

  { TODO: Now using optimised bubble sort (faster than Quick-sort if there are
    few elements unordered).  It does only one iteration, because sprites move
    only few pixels each frames.  Also it orders from last to first, because
    new sprites are inserted at the end of the list, so it moves to its place
    faster  (NOTE:  actually I'm not sure about the last sentence.  I need to
    evaluate it).

    May be it's possible to optimise this or use another algorightm (Ping-pong
    is good candidate).  Needs to be profiled. }
    FOR Ndx := LOW (SpriteOrder) TO (HIGH (SpriteOrder) - 1) DO
      IF SpriteList[SpriteOrder[Ndx]].Y < SpriteList[SpriteOrder[Ndx + 1]].Y THEN
      BEGIN
	Tmp := SpriteOrder[Ndx];
	SpriteOrder[Ndx] := SpriteOrder[Ndx + 1];
	SpriteOrder[Ndx + 1] := Tmp;
      END;
  { This is used by the PlaySoundSample procedure to calculate stereo panning. }
    LastScrollX := ScrollX;
  { Draw the sprites. }
    FOR Ndx := High (SpriteOrder) DOWNTO Low (SpriteOrder) DO
    BEGIN
    { Draw only active sprites. }
      IF (SpriteList[SpriteOrder[Ndx]].Y <> SPR_Y_OUT)
      AND (SpriteList[SpriteOrder[Ndx]].Picture <> NIL) THEN
      BEGIN
      { Calculate screen coordinates from world ones. }
	Posx := SpriteList[SpriteOrder[Ndx]].x - ScrollX + SpriteList[SpriteOrder[Ndx]].OffX;
	Posy := SpriteList[SpriteOrder[Ndx]].y - ScrollY + SpriteList[SpriteOrder[Ndx]].OffY;
      { Get the RLE description. }
	RLESprite := SpriteList[SpriteOrder[Ndx]].Picture;
      { Check if the sprite is inside the output bitmap. }
	IF ((-RLESprite^.w) < Posx) AND ((-RLESprite^.h) < Posy)
	AND (Posx < Bmp^.w) AND (Posy < Bmp^.h) THEN
	{ Draw the sprite. }
	  al_draw_rle_sprite (Bmp, RLESprite, PosX, PosY);
      END;
    END;
  END;



(* Draws the sprites using the lightmap from unit "Lighting" to light them. *)
  PROCEDURE DrawLighted (Bmp: AL_BITMAPptr; ScrollX, ScrollY: INTEGER);
  VAR
    Ndx, Tmp: INTEGER;
    Posx, Posy: INTEGER;
    RLESprite: AL_RLE_SPRITEptr;
    LightColor: INTEGER;
  BEGIN
  { Order sprites by Y.  Before you ask about the ordering algorithm, please
    read next paragraph:

    The "Quick-sort" is the fastest in the worst case (i.e. the list is fully
    unordered) but if the list is almost ordered (as the sprite list is) it is
    quite slow actually.  That's why I'm using a different one here. }

  { TODO: Now using optimised bubble sort (faster than Quick-sort if there are
    few elements unordered).  It does only one iteration, because sprites move
    only few pixels each frames.  Also it orders from last to first, because
    new sprites are inserted at the end of the list, so it moves to its place
    faster  (NOTE:  actually I'm not sure about the last sentence.  I need to
    evaluate it).

    May be it's possible to optimise this or use another algorightm (Ping-pong
    is good candidate).  Needs to be profiled. }
    FOR Ndx := LOW (SpriteOrder) TO (HIGH (SpriteOrder) - 1) DO
      IF SpriteList[SpriteOrder[Ndx]].Y < SpriteList[SpriteOrder[Ndx + 1]].Y THEN
      BEGIN
	Tmp := SpriteOrder[Ndx];
	SpriteOrder[Ndx] := SpriteOrder[Ndx + 1];
	SpriteOrder[Ndx + 1] := Tmp;
      END;
  { This is used by the PlaySoundSample procedure to calculate stereo panning. }
    LastScrollX := ScrollX;
  { Draw the sprites. }
    FOR Ndx := High (SpriteOrder) DOWNTO Low (SpriteOrder) DO
    BEGIN
    { Draw only active sprites. }
      IF (SpriteList[SpriteOrder[Ndx]].Y <> SPR_Y_OUT)
      AND (SpriteList[SpriteOrder[Ndx]].Picture <> NIL) THEN
      BEGIN
      { Calculate screen coordinates from world ones. }
	Posx := SpriteList[SpriteOrder[Ndx]].x - ScrollX + SpriteList[SpriteOrder[Ndx]].OffX;
	Posy := SpriteList[SpriteOrder[Ndx]].y - ScrollY + SpriteList[SpriteOrder[Ndx]].OffY;
      { Get the RLE description. }
	RLESprite := SpriteList[SpriteOrder[Ndx]].Picture;
      { Check if the sprite is inside the output bitmap. }
	IF ((-RLESprite^.w) < Posx) AND ((-RLESprite^.h) < Posy)
	AND (Posx < Bmp^.w) AND (Posy < Bmp^.h) THEN
	BEGIN
	{ Get light.  It needs only the alpha channel. }
	  LightColor := al_getpixel (LightMap,
	    SpriteList[SpriteOrder[Ndx]].x - ScrollX + 32,
	    SpriteList[SpriteOrder[Ndx]].y - ScrollY + 64) SHR 24;
	{ Draw the sprite. }
	  al_draw_lit_rle_sprite (Bmp, RLESprite, PosX, PosY, LightColor);
	END;
      END;
    END;
  END;



(* Plays a sound sample changing the stereo 'pan' as the X coordinate of the
   sprite. *)
  PROCEDURE PlaySoundSample (Spr: INTEGER; Sample: AL_SAMPLEptr);
  VAR
    Pan: INTEGER;
  BEGIN
  { Calculate the 'pan'. }
    Pan := (SpriteList[Spr].x - LastScrollX) * 256 DIV 640;
  { Play the sample. }
    al_play_sample (Sample, 127, Pan, 1000, 0);
  END;

END.
