UNIT tilemap;
(* Program: Duke of Dragonfear: Deathland.
 * Description: The tilemap engine.
 * Author: Ñuño Martínez <niunio at users.sourceforge.net>
 *	   Translated from a game by Ken Silverman <http://www.advsys.net/ken/>
 *)
(*
  Copyright (c) 2011 Guillermo Martínez J.

  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software
    in a product, an acknowledgment in the product documentation would be
    appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
    distribution.
*)

INTERFACE

USES
  allegro; { Bitmap manipulation. }

CONST
(* Map size. *)
  Width = 66;
  Height = Width;
(* Size of tiles in pixels. *)
  TSIZE = 64;
VAR
(* The tilemap. *)
  Board: ARRAY [1..Width, 1..Height] OF BYTE;



(* Gets the tile bitmaps from the given tile.
 * The bitmap should be 512x512 pixels, where each tile is 64x64 pixels. *)
  PROCEDURE GetTilesetFromBitmap (CONST Bmp: AL_BITMAPptr);

(* Returns TRUE if tile is walkable. *)
  FUNCTION IsWalkable (CONST X, Y: INTEGER): BOOLEAN; INLINE;

(* Returns TRUE if tile is wall. *)
  FUNCTION IsWall (CONST X, Y: INTEGER): BOOLEAN; INLINE;

(* Be sure that the scroll isn't out of the edges of the board.  Should be
   used before draw anything. *)
  PROCEDURE FixScroll (CONST Bmp: AL_BITMAPptr; CONST Ix, Iy: INTEGER;
			OUT Ox, Oy: INTEGER); INLINE;

(* Draws the board in the given bitmap at the given scroll coordinates. *)
  PROCEDURE Draw (Bmp: AL_BITMAPptr; ScrollX, ScrollY: INTEGER);

(* Release tileset. *)
  PROCEDURE DestroyTileset;



IMPLEMENTATION

USES
  sysutils; { For string manipulation. }


VAR
  fTileSet: ARRAY [0..63] OF AL_BITMAPptr;



(* Gets the tile bitmaps from the given tile.
 * The bitmap should be 512x512 pixels, where each tile is 64x64 pixels. *)
  PROCEDURE GetTilesetFromBitmap (CONST Bmp: AL_BITMAPptr);
  VAR
    Ndx, X, Y: INTEGER;
  BEGIN
    IF (Bmp^.w < TSIZE * 8) OR (Bmp^.h < TSIZE * 8) THEN
      RAISE Exception.Create ('Tile set bitmap size isn''t correct.');
    DestroyTileset;
    Ndx := 0;
    FOR Y := 0 TO 7 DO
      FOR X := 0 TO 7 DO
      BEGIN
      { TODO: Use sub-bitmaps! }
	fTileSet[Ndx] := al_create_bitmap (TSIZE, TSIZE);
	al_blit (Bmp, fTileSet[Ndx],
	  X * TSIZE, Y * TSIZE, 0, 0,
	  TSIZE, TSIZE);
	INC (Ndx);
      END;
  END;



(* Returns TRUE if tile is walkable. *)
  FUNCTION IsWalkable (CONST X, Y: INTEGER): BOOLEAN;
  BEGIN
    IsWalkable := Board[X, Y] IN
  { The position of number is the same at the tileset. }
    [  0,  1,  2,  3,  4,          7,
                                  15,
                                  23,
                              30, 31,
          33, 34, 35, 36,     38,
          41, 42,     44,     46,
          49,     51, 52,     54,
                              62, 63];
  END;



(* Returns TRUE if tile is wall. *)
  FUNCTION IsWall (CONST X, Y: INTEGER): BOOLEAN;
  BEGIN
    IsWall := Board[X, Y] IN
  { The position of number is the same at the tileset. }
    [                      5, 6,
       8,  9, 10, 11, 12, 13, 14,
      16, 17, 18, 19, 20, 21, 22,
      24, 25, 26, 27, 28, 29,
      32,                 37,     39,
      40,         43,     45,     47,
      48,     50,         53,     55,
      56, 57, 58, 59, 60, 61        ];
  END;



(* Be sure that the scroll isn't out of the edges of the board.  Should be
   used before to draw anything. *)
  PROCEDURE FixScroll (CONST Bmp: AL_BITMAPptr; CONST Ix, Iy: INTEGER;
			OUT Ox, Oy: INTEGER);
  BEGIN
    IF Ix < 0 THEN
      Ox := 0
    ELSE IF Ix > (Width * TSIZE) - Bmp^.w THEN
      Ox := (Width * TSIZE) - Bmp^.w
    ELSE
      Ox := Ix;

    IF Iy < 0 THEN
      Oy := 0
    ELSE IF Iy > (Height * TSIZE) - Bmp^.h THEN
      Oy := (Height * TSIZE) - Bmp^.h
    ELSE
      Oy := Iy;
  END;



(* Draws the board in the given bitmap at the given scroll coordinates.
 * I'm sure it can be optimized a lot (and it should be) but I try to keep
 * it simple and understandable. *)
  PROCEDURE Draw (Bmp: AL_BITMAPptr; ScrollX, ScrollY: INTEGER);
  VAR
    NumTilesW, NumTilesH: INTEGER; { Number of tiles to draw. }
    FirstTileX, FirstTileY: INTEGER; { First tile to draw. }
    OffsetX, OffsetY: INTEGER; { Relative coordinates of first tile. }
    PosX, PosY: INTEGER; { Where to draw the tile. }
    X, Y: INTEGER; { Tile to draw. }
  BEGIN
  { Calculate how many tiles can be drawn in the bitmap. }
    NumTilesW  := (Bmp^.w DIV TSIZE) + 1; { Adds one for the edges. }
    NumTilesH  := (Bmp^.h DIV TSIZE) + 1;
    IF NumTilesW > Width THEN
      NumTilesW := Width;
    IF NumTilesH > Height THEN
      NumTilesH := Height;
  { Calculate the first tile to be drawn. }
    FirstTileX := (ScrollX DIV TSIZE) + 1;
    FirstTileY := (ScrollY DIV TSIZE) + 1;
  { Calculate the offset of that first tile.  They are the coordinates where the
    first tile will be drawn.  The tile offset is the pixel of the tile that is in
    the upper left pixel of the output bitmap. }
    OffsetX := -(ScrollX MOD TSIZE);
    OffsetY := -(ScrollY MOD TSIZE);
  { Draw. }
    PosY := OffsetY;
    FOR Y := FirstTileY TO FirstTileY + NumTilesH DO
    BEGIN
      PosX := OffsetX;
      FOR X := FirstTileX TO FirstTileX + NumTilesW DO
      BEGIN
	IF (X <= Width) AND (Y <= Height) THEN
	  al_blit (fTileSet[Board[X, Y]], Bmp, 0, 0, PosX, PosY, TSIZE, TSIZE);
      { Increment the tile position. }
	INC (PosX, TSIZE);
      END;
    { Increment the tile position. }
      INC (PosY, TSIZE);
    END;
  END;



(* Release tileset. *)
  PROCEDURE DestroyTileset;
  VAR
    Ndx: INTEGER;
  BEGIN
    FOR Ndx := LOW (fTileSet) TO HIGH (fTileSet) DO
      IF fTileSet[Ndx] <> NIL THEN
	al_destroy_bitmap (fTileSet[Ndx]);
  END;

END.
