UNIT usrInput;
(*<Implements the user input system.

TODO: It also implements the input configuration dialog.

  Note the input implementation is a fast implementation.  May be there's a
  more efficient way to do this stuff. *)
(* Copyright (c) 2011 Guillermo Martínez J.

  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software
    in a product, an acknowledgment in the product documentation would be
    appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
    distribution.
 *)

INTERFACE

CONST
(* Mnemonics for lever direction.
   Note that they coincide with pictures! *)
  DIR_NOP = 0;
  DIR_UP = 3;
  DIR_UP_RIGHT = 4;
  DIR_RIGHT = 5;
  DIR_DOWN_RIGHT = 6;
  DIR_DOWN = 7;
  DIR_DOWN_LEFT = 8;
  DIR_LEFT = 1;
  DIR_UP_LEFT = 2;

(* Direction groups. *)
  DIR_G_UP    = [DIR_UP_LEFT, DIR_UP, DIR_UP_RIGHT];
  DIR_G_DOWN  = [DIR_DOWN_LEFT, DIR_DOWN, DIR_DOWN_RIGHT];
  DIR_G_RIGHT = [DIR_UP_RIGHT, DIR_RIGHT, DIR_DOWN_RIGHT];
  DIR_G_LEFT  = [DIR_UP_LEFT, DIR_LEFT, DIR_DOWN_LEFT];

(* Mnemonics for buttons. *)
  BTN_ATTACK = 0;
  BTN_Z = 1;

TYPE
(* Interface for user input.  Should load the configuration from the
   configuration file in the constructor. *)
  TUserInput = CLASS (TObject)
  PUBLIC
  (* Constructor.  Should load the configuration from the configuration file.
     @param(PlayerNumber 1 or 2, to identify it and know what configuration to
     use.) *)
     CONSTRUCTOR Create (PlayerNumber: INTEGER); VIRTUAL;
  (* Returns the direction of the "lever".  See mnemonics above. *)
    FUNCTION LeverDir: INTEGER; VIRTUAL; ABSTRACT;
  (* Should return @true if button is pressed.
    @param(aButton Button identifier.) *)
    FUNCTION Button (aButton: INTEGER): BOOLEAN; VIRTUAL; ABSTRACT;
  (* Returns @true if any button is pressed. *)
    FUNCTION AnyButton: BOOLEAN; INLINE;
  END;



(* Implements keyboard input. *)
  TKeyboardInput = CLASS (TUserInput)
  PRIVATE
    fUpScanCode, fRightScanCode, fDownScanCode, fLeftScanCode: INTEGER;
    fBtnScanCode: ARRAY [BTN_ATTACK .. BTN_Z] OF INTEGER;

  (* Returns the direction scan-code key. *)
    FUNCTION GetDirScanCode (aDirection: INTEGER): INTEGER;
  (* Sets the direction scan-code key. *)
    PROCEDURE SetDirScanCode (aDirection, aScanCode: INTEGER);
  (* Returns the button scan-code key. *)
    FUNCTION GetBtnScanCode (aButton: INTEGER): INTEGER;
  (* Sets the button scan-code key. *)
    PROCEDURE SetBtnScanCode (aButton, aScanCode: INTEGER);
  PUBLIC
  (* Constructor.  Loads the configuration from the configuration file.
     @param(PlayerNumber 1 or 2, to identify it and know what configuration to
     use.) *)
     CONSTRUCTOR Create (PlayerNumber: INTEGER); OVERRIDE;
  (* Returns the direction of the "lever".  See mnemonics above. *)
    FUNCTION LeverDir: INTEGER; OVERRIDE;
  (* Returns @true if button is pressed.
    @param(aButton Button identifier.) *)
    FUNCTION Button (aButton: INTEGER): BOOLEAN; OVERRIDE;
  (* To access to the direction scan-codes. *)
    PROPERTY DirScanCode[aDirection: INTEGER]: INTEGER READ GetDirScanCode WRITE SetDirScanCode;
  (* To access to the button scan-codes. *)
    PROPERTY BtnScanCode[aButton: INTEGER]: INTEGER READ GetBtnScanCode WRITE SetBtnScanCode;
  END;



(* Implements joystick input. *)
  TJoystickInput = CLASS (TUserInput)
  PRIVATE
    fButtonDict: ARRAY [BTN_ATTACK .. BTN_Z] OF INTEGER;
  PUBLIC
  (* Constructor.  Loads the configuration from the configuration file.
     @param(PlayerNumber 1 or 2, to identify it and know what configuration to
     use.) *)
     CONSTRUCTOR Create (PlayerNumber: INTEGER); OVERRIDE;
  (* Returns the direction of the "lever".  See mnemonics above. *)
    FUNCTION LeverDir: INTEGER; OVERRIDE;
  (* Returns @true if button is pressed.
    @param(aButton Button identifier.) *)
    FUNCTION Button (aButton: INTEGER): BOOLEAN; OVERRIDE;
  END;

VAR
(* References to the user input.  They're created by @link(config.GetConfiguration). *)
  UserInput: ARRAY [1..2] OF TUserInput;



(* Entry to the configuration dialog. *)
  PROCEDURE ConfigurationDialog;

IMPLEMENTATION

USES
  allegro,
  sysutils;



(**************
 * TUserInput *
 **************)

(* Constructor.  Should load the configuration from the configuration file.
   @param(PlayerNumber 1 or 2, to identify it and know what configuration to
   use.) *)
  CONSTRUCTOR TUserInput.Create (PlayerNumber: INTEGER);
  BEGIN
    IF NOT (PlayerNumber IN [1, 2]) THEN
      RAISE Exception.Create ('Wrong player identifier while initializing player input.');
    INHERITED Create;
  END;



(* Returns @true if any button is pressed. *)
  FUNCTION TUserInput.AnyButton: BOOLEAN;
  BEGIN
    AnyButton :=   SELF.Button (BTN_ATTACK);
  END;



(******************
 * TKeyboardInput *
 ******************)

(* Returns the direction scan-code key. *)
  FUNCTION TKeyboardInput.GetDirScanCode (aDirection: INTEGER): INTEGER;
  BEGIN
    CASE aDirection OF
      DIR_UP:
	GetDirScanCode := fUpScanCode;
      DIR_DOWN:
	GetDirScanCode := fDownScanCode;
      DIR_RIGHT:
	GetDirScanCode := fRightScanCode;
      DIR_LEFT:
	GetDirScanCode := fLeftScanCode;
      ELSE
	RAISE Exception.Create ('Requested direction hasn''t scan-code.');
      END;
  END;



(* Sets the direction scan-code key. *)
  PROCEDURE TKeyboardInput.SetDirScanCode (aDirection, aScanCode: INTEGER);
  BEGIN
    CASE aDirection OF
      DIR_UP:
	fUpScanCode := aScanCode;
      DIR_DOWN:
	fDownScanCode := aScanCode;
      DIR_RIGHT:
	fRightScanCode := aScanCode;
      DIR_LEFT:
	fLeftScanCode := aScanCode;
      ELSE
	RAISE Exception.Create ('Requested direction hasn''t scan-code.');
      END;
  END;



(* Returns the button scan-code key. *)
  FUNCTION TKeyboardInput.GetBtnScanCode (aButton: INTEGER): INTEGER;
  BEGIN
    GetBtnScanCode := fBtnScanCode[aButton];
  END;



(* Sets the button scan-code key. *)
  PROCEDURE TKeyboardInput.SetBtnScanCode (aButton, aScanCode: INTEGER);
  BEGIN
    fBtnScanCode[aButton] := aScanCode;
  END;




CONST
(* Scan-codes defaults for keyboard. *)
  KEY_UP   : ARRAY[1..2] OF INTEGER = (AL_KEY_W, AL_KEY_UP);
  KEY_DOWN : ARRAY[1..2] OF INTEGER = (AL_KEY_S, AL_KEY_DOWN);
  KEY_RIGHT: ARRAY[1..2] OF INTEGER = (AL_KEY_D, AL_KEY_RIGHT);
  KEY_LEFT : ARRAY[1..2] OF INTEGER = (AL_KEY_A, AL_KEY_LEFT);

  KEY_ATTACK  : ARRAY[1..2] OF INTEGER = (AL_KEY_T, AL_KEY_7_PAD);



(* Constructor.  Loads the configuration from the configuration file.
   @param(PlayerNumber 1 or 2, to identify it and know what configuration to
   use.) *)
  CONSTRUCTOR TKeyboardInput.Create (PlayerNumber: INTEGER);
  VAR
    PlayerSectionName: STRING;
  BEGIN
    INHERITED Create (PlayerNumber);
    PlayerSectionName := 'player' + IntToStr (PlayerNumber);
    fUpScanCode    := al_get_config_int (PlayerSectionName,'up',    KEY_UP   [PlayerNumber]);
    fRightScanCode := al_get_config_int (PlayerSectionName,'right', KEY_RIGHT[PlayerNumber]);
    fDownScanCode  := al_get_config_int (PlayerSectionName,'down',  KEY_DOWN [PlayerNumber]);
    fLeftScanCode  := al_get_config_int (PlayerSectionName,'left',  KEY_LEFT [PlayerNumber]);

    fBtnScanCode[BTN_ATTACK] := al_get_config_int (PlayerSectionName, 'attack',   KEY_ATTACK [PlayerNumber]);
  END;



(* Returns the direction of the "lever".  See mnemonics above. *)
  FUNCTION TKeyboardInput.LeverDir: INTEGER;
  BEGIN
    IF al_keyboard_needs_poll THEN
      al_poll_keyboard;
    IF al_key[SELF.fUpScanCode] <> 0 THEN
    BEGIN
      IF al_key[SELF.fLeftScanCode] <> 0 THEN
	LeverDir := DIR_UP_LEFT
      ELSE IF al_key[SELF.fRightScanCode] <> 0 THEN
	LeverDir := DIR_UP_RIGHT
      ELSE
	LeverDir := DIR_UP;
    END
    ELSE IF al_key[SELF.fDownScanCode] <> 0 THEN
    BEGIN
      IF al_key[SELF.fLeftScanCode] <> 0 THEN
	LeverDir := DIR_DOWN_LEFT
      ELSE IF al_key[SELF.fRightScanCode] <> 0 THEN
	LeverDir := DIR_DOWN_RIGHT
      ELSE
	LeverDir := DIR_DOWN;
    END
    ELSE IF al_key[SELF.fLeftScanCode] <> 0 THEN
      LeverDir := DIR_LEFT
    ELSE IF al_key[SELF.fRightScanCode] <> 0 THEN
      LeverDir := DIR_RIGHT
    ELSE
      LeverDir := DIR_NOP;
  END;



(* Returns @true if button is pressed.
  @param(aButton Button identifier.) *)
  FUNCTION TKeyboardInput.Button (aButton: INTEGER): BOOLEAN;
  BEGIN
    Button := al_key[SELF.fBtnScanCode[aButton]] <> 0;
  END;



(******************
 * TJoystickInput *
 ******************)

(* Constructor.  Loads the configuration from the configuration file.
   @param(PlayerNumber 1 or 2, to identify it and know what configuration to
   use.) *)
  CONSTRUCTOR TJoystickInput.Create (PlayerNumber: INTEGER);
  VAR
    PlayerSectionName: STRING;
    Ndx: INTEGER;
  CONST
    BtnName: ARRAY [BTN_ATTACK .. BTN_Z] OF STRING = (
      'attack', 'extra');
  BEGIN
    INHERITED Create (PlayerNumber);
    PlayerSectionName := 'player' + IntToStr (PlayerNumber);
    FOR Ndx := Low (fButtonDict) TO High (fButtonDict) DO
      fButtonDict[Ndx] := al_get_config_int (PlayerSectionName, BtnName[Ndx],   Ndx);
  END;



(* Returns the direction of the "lever".  See mnemonics above. *)
  FUNCTION TJoystickInput.LeverDir: INTEGER;
  BEGIN
    LeverDir := DIR_NOP;
  END;



(* Returns @true if button is pressed.
  @param(aButton Button identifier.) *)
  FUNCTION TJoystickInput.Button (aButton: INTEGER): BOOLEAN;
  BEGIN
    Button := FALSE;
  END;



(**************************************
 * Entry to the configuration dialog. *
 **************************************)
  PROCEDURE ConfigurationDialog;
  BEGIN
    RAISE Exception.Create ('Input configuration dialog not implemented.');
  END;

END.
